# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Compatibility layer of Hugging Face datasets library."""

from __future__ import annotations

import builtins
import contextlib
import functools
import glob
import os
import sys
import types
from typing import Iterator, NamedTuple, Optional, Union
from unittest import mock

from etils import epath
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import download
from tensorflow_datasets.core import features
from tensorflow_datasets.core import logging as tfds_logging
from tensorflow_datasets.core import split_builder
from tensorflow_datasets.core import splits
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

__all__ = [
    'mock_builtin_to_use_gfile',
    'mock_huggingface_import',
]


@contextlib.contextmanager
def mock_huggingface_import() -> Iterator[None]:
  """Patch the HuggingFace datasets module to use TFDS.

  This allow using datasets defined in HG directly in TFDS.

  Usage:

  ```python
  import tensorflow_datasets as tfds

  with tfds.core.community.mock_huggingface_import():
    import datasets  # `datasets` is a _MockedHFDatasets

  # Using `datasets.Xyz` uses the corresponding `tfds.Xyz` API
  class MyDataset(datasets.GeneratorBasedBuilder):
    version = datasets.Version('1.0.0')
    ...

  # This works !!
  ds = tfds.load('my_dataset')
  ```

  Yields:
    None
  """
  # Replace `datasets` module by the TFDS mock
  old_hg_module = sys.modules.get('datasets')
  try:
    sys.modules['datasets'] = _MockedHFDatasets()
    yield
  finally:
    if old_hg_module:
      sys.modules['datasets'] = old_hg_module
    else:
      del sys.modules['datasets']


# Wrappers around `DatasetBuilder` & `DatasetInfo`


class GeneratorBasedBuilder(
    dataset_builder.GeneratorBasedBuilder,
    skip_registration=True,
):
  """Mocked Generator builder."""

  VERSION = utils.Version('1.0.0')

  # TODO(tfds): Support manual_data
  # MANUAL_DOWNLOAD_INSTRUCTIONS = 'Missing instructions.'

  BUILDER_CONFIG_CLASS = dataset_builder.BuilderConfig
  # In HF, default config is explicitly defined
  DEFAULT_CONFIG_NAME = None

  def __init_subclass__(cls, **kwargs):
    # In HF, default config is explicitly defined
    if cls.DEFAULT_CONFIG_NAME:
      if not cls.BUILDER_CONFIGS:
        # If no config is defined, HF implicitly create a
        # default config (e.g. `clickbait_news_bg`)
        cls.BUILDER_CONFIGS = [
            cls.BUILDER_CONFIG_CLASS(name=cls.DEFAULT_CONFIG_NAME)
        ]
      else:
        # Should re-order config so that the first config
        # is the default one
        name_to_configs = {c.name: c for c in cls.BUILDER_CONFIGS}
        new_ordered_configs = [name_to_configs[cls.DEFAULT_CONFIG_NAME]]
        new_ordered_configs.extend(
            config
            for name, config in name_to_configs.items()
            if name != cls.DEFAULT_CONFIG_NAME
        )
        cls.BUILDER_CONFIGS = new_ordered_configs
    super().__init_subclass__(**kwargs)

  @property
  @tfds_logging.builder_info()
  @functools.lru_cache(maxsize=128)
  def info(self) -> dataset_info.DatasetInfo:
    # HF supports Sequences defined as list: {'video': [{'image': Image()}]}
    with _mock_list_as_sequence():
      info = self._info()
    # HF DatasetInfo do not have `builder` args, so we insert
    # here
    return dataset_info.DatasetInfo(builder=self, **info.kwargs)

  @property
  def config(self) -> dataset_builder.BuilderConfig:
    # HF datasets uses `self.config` rather than `self.builder_config`
    return self.builder_config

  def _download_and_prepare(self, *args, **kwargs):
    # * Patch `open` to use the GFile API (to supports GCS)
    # * Patch `Tensor.encode_example` to support `None` strings
    with mock_builtin_to_use_gfile(), _mock_tensor_feature_empty_str():
      return super()._download_and_prepare(*args, **kwargs)


class BeamBasedBuilder(
    GeneratorBasedBuilder,
    dataset_builder.BeamBasedBuilder,
    skip_registration=True,
):
  pass


class DatasetInfo:
  """Wrapper around `DatasetInfo` kwargs."""

  def __init__(self, **kwargs):  # pylint: disable=redefined-builtin
    self.kwargs = kwargs


# Wrapper around the tfds.features


def _make_scalar_feature(dtype: str) -> tf.dtypes.DType:
  """Returns the `tf.dtype` scalar feature."""
  str2val = {
      'bool_': tf.bool,
      'float': tf.float32,
      'double': tf.float64,
      'large_string': tf.string,
      'utf8': tf.string,
  }
  if dtype in str2val:
    return str2val[dtype]
  elif hasattr(tf.dtypes, dtype):
    return getattr(tf.dtypes, dtype)
  else:
    raise ValueError(
        f'Unrecognized type {dtype}. Please open an issue if you think '
        'this is a bug.'
    )


class Value:

  def __new__(cls, dtype):
    return _make_scalar_feature(dtype)


@contextlib.contextmanager
def _mock_tensor_feature_empty_str() -> Iterator[None]:
  """Support `yield {'text': None}`."""

  # Note: We cannot create some custom
  # `class Text(tfds.features.Tensor)` as it would be
  # badly exported when reading from files.

  encode_example_fn = features.Tensor.encode_example

  def new_encode_example(self, example_data):
    if example_data is None and self.dtype == tf.string:
      example_data = ''
    return encode_example_fn(self, example_data)

  with mock.patch.object(features.Tensor, 'encode_example', new_encode_example):
    yield


@contextlib.contextmanager
def _mock_list_as_sequence() -> Iterator[None]:
  """Supports `Sequence` defined as list (e.g. `{'video': [{'frame': ...}]`."""

  to_feature_fn = features.features_dict.to_feature

  def new_to_feature(value):
    if isinstance(value, list):
      (value,) = (
          value  # List should contain a single element  # pylint: disable=self-assigning-variable
      )
      return features.Sequence(value)
    else:
      return to_feature_fn(value)

  # Should this be supported nativelly ?
  with mock.patch.object(features.features_dict, 'to_feature', new_to_feature):
    yield


# Other wrappers


class Version(utils.Version):

  def __init__(
      self,
      version: Union[utils.Version, str],
      description: Optional[str] = None,
  ):
    del description  # Description not used.
    super().__init__(version)
    self.version_str = str(self)  # e.g. used by `compguesswhat`


class _MockedHFDatasets(types.ModuleType):
  """Hugging face API mocked to redirect to TFDS methods."""

  def __init__(self):
    super().__init__('datasets')

  # pylint: disable=invalid-name

  Version = Version
  Split = splits.Split
  SplitGenerator = split_builder.SplitGeneratorLegacy
  DownloadManager = download.DownloadManager

  # dataset_builder.py
  GeneratorBasedBuilder = GeneratorBasedBuilder
  BeamBasedBuilder = BeamBasedBuilder
  BuilderConfig = dataset_builder.BuilderConfig
  DatasetInfo = DatasetInfo

  class features(types.ModuleType):
    """`datasets.features` module."""

    Features = features.FeaturesDict
    Sequence = features.Sequence
    ClassLabel = features.ClassLabel
    Value = Value
    Translation = features.Translation

  class info(types.ModuleType):
    """`datasets.info` module."""

    class SupervisedKeysData(NamedTuple):  # pylint: disable=invalid-namedtuple-arg
      input: str
      output: str

  # pylint: enable=invalid-name


# `datasets.features.Xyz` API can also be accessed as `datasets.Xyz`
for attr in dir(_MockedHFDatasets.features):
  if attr.startswith('_'):
    continue
  setattr(_MockedHFDatasets, attr, getattr(_MockedHFDatasets.features, attr))

# Mock the standard file API to use GFile


def _new_open(*args, encoding=None, **kwargs):
  """Mocked `open`."""
  if encoding and encoding.lower() not in ('utf8', 'utf-8'):
    raise ValueError('Do not support non UTF-8 encoding')
  return tf.io.gfile.GFile(*args, **kwargs)


def _new_pathlib_path_new(cls, *args):
  """Mocked `pathlib.Path.__new__`."""
  del cls
  return epath.Path(*args)


@contextlib.contextmanager
def mock_builtin_to_use_gfile() -> Iterator[None]:
  """Mock the standard library to support GCS access.

  Usage:

  ```python
  with tfds.core.community.mock_builtin_to_use_gfile():
    files = os.listdir('gs://some-bucket')
  ```

  Yields:
    None
  """
  with mock.patch.object(builtins, 'open', _new_open), mock.patch(
      # Patch `__new__` as `Path` might have already be imported as
      # `from pathlib import Path`
      'pathlib.Path.__new__',
      _new_pathlib_path_new,
  ), mock.patch.object(
      os.path, 'exists', tf.io.gfile.exists
  ), mock.patch.object(
      os.path, 'isdir', tf.io.gfile.isdir
  ), mock.patch.object(
      os, 'listdir', tf.io.gfile.listdir
  ), mock.patch.object(
      os, 'walk', tf.io.gfile.walk
  ), mock.patch.object(
      glob, 'glob', tf.io.gfile.glob
  ):
    yield



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Beam pipeline which computes the number of examples in the given tfrecord.

Compute the split info (num shards, num examples,...) metadata required
by `tfds.core.DatasetInfo`. See documentation and usage at:
https://www.tensorflow.org/datasets/external_tfrecord

"""

import argparse
from typing import List

from absl import app
from absl.flags import argparse_flags
import tensorflow_datasets as tfds
from tensorflow_datasets.scripts.utils import flag_utils


# Open source pytests globally import all files, so create duplicate FLAGS
# error when using absl.flags. So we use argparse instead.
def _parse_flags(argv: List[str]) -> argparse.Namespace:
  """Parses command line flags."""
  argv = flag_utils.normalize_flags(argv)  # See b/174043007 for context.

  parser = argparse_flags.ArgumentParser(
      description='Tensorflow Datasets CLI tool',
  )
  parser.add_argument(
      '--data_dir',
      type=tfds.core.Path,
      help='Path to the dataset files.',
  )
  parser.add_argument(
      '--out_dir',
      type=tfds.core.Path,
      help='Computed metadata will be written here.',
  )
  return parser.parse_args(argv[1:])


def main(args: argparse.Namespace) -> None:

  tfds.folder_dataset.compute_split_info_from_directory(
      data_dir=args.data_dir,
      out_dir=args.out_dir,
  )


if __name__ == '__main__':
  app.run(main, flags_parser=_parse_flags)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests tensorflow_datasets.core.naming."""

import os
from absl.testing import parameterized
from etils import epath
import pytest
from tensorflow_datasets import testing
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import splits

_FILENAME_TEMPLATE_DEFAULT = naming.ShardedFileTemplate(data_dir='.')
_FILENAME_TEMPLATE_MNIST_DEFAULT = naming.ShardedFileTemplate(
    data_dir='/path',
    dataset_name='mnist',
    split='train',
    filetype_suffix='tfrecord',
)
_FILENAME_TEMPLATE_CUSTOM_NO_EXTRA = naming.ShardedFileTemplate(
    data_dir='/path', template='{SPLIT}.{SHARD_INDEX}'
)
_FILENAME_TEMPLATE_CUSTOM_FULL = naming.ShardedFileTemplate(
    data_dir='/path',
    template='{SPLIT}.{SHARD_INDEX}',
    dataset_name='mnist',
    split='train',
    filetype_suffix='tfrecord',
)


class NamingTest(parameterized.TestCase, testing.TestCase):

  @parameterized.parameters(
      ('HelloWorld', 'hello_world'),
      ('FooBARBaz', 'foo_bar_baz'),
      ('FooBar123', 'foo_bar123'),
      ('FooBar123Baz', 'foo_bar123_baz'),
      ('FooBar123baz', 'foo_bar123baz'),
  )
  def test_camelcase_to_snakecase(self, camel, snake):
    self.assertEqual(snake, naming.camelcase_to_snakecase(camel))

  @parameterized.parameters(
      ('HelloWorld', 'hello_world'),
      ('FooBar123', 'foo_bar123'),
      ('FooBar123Baz', 'foo_bar123_baz'),
      ('FooBar123baz', 'foo_bar123baz'),
  )
  def test_snake_to_camelcase(self, camel, snake):
    self.assertEqual(naming.snake_to_camelcase(snake), camel)
    # camelcase_to_snakecase is a no-op if the name is already snake_case.
    self.assertEqual(naming.camelcase_to_snakecase(snake), snake)

  @parameterized.parameters(
      (2, '00000-of-00002', '00001-of-00002'),
      (12345, '00000-of-12345', '12344-of-12345'),
      (654321, '000000-of-654321', '654320-of-654321'),
  )
  def test_sharded_filepaths(
      self,
      num_shards: int,
      expected_first_suffix: str,
      expected_last_suffix: str,
  ):
    template = naming.ShardedFileTemplate(
        split='train',
        dataset_name='ds',
        data_dir='/path',
        filetype_suffix='tfrecord',
    )
    names = template.sharded_filepaths(num_shards)
    path_template = '/path/ds-train.tfrecord-%s'
    self.assertEqual(
        os.fspath(names[0]), path_template % (expected_first_suffix)
    )
    self.assertEqual(
        os.fspath(names[-1]), path_template % (expected_last_suffix)
    )

  @parameterized.parameters(
      ('foo', 'foo-train'),
      ('Foo', 'foo-train'),
      ('FooBar', 'foo_bar-train'),
  )
  def test_filename_prefix_for_split(self, prefix, expected):
    split = splits.Split.TRAIN
    self.assertEqual(expected, naming.filename_prefix_for_split(prefix, split))

  def test_filenames_for_dataset_split(self):
    actual = naming.filenames_for_dataset_split(
        dataset_name='foo',
        split=splits.Split.TRAIN,
        num_shards=2,
        filetype_suffix='bar',
        data_dir='/path',
    )
    self.assertEqual(
        actual, ['foo-train.bar-00000-of-00002', 'foo-train.bar-00001-of-00002']
    )

  def test_filepaths_for_dataset_split(self):
    actual = naming.filepaths_for_dataset_split(
        dataset_name='foo',
        split=splits.Split.TRAIN,
        num_shards=2,
        data_dir='/tmp/bar/',
        filetype_suffix='bar',
    )
    self.assertEqual(
        actual,
        [
            '/tmp/bar/foo-train.bar-00000-of-00002',
            '/tmp/bar/foo-train.bar-00001-of-00002',
        ],
    )

  def test_filepattern_for_dataset_split(self):
    self.assertEqual(
        '/tmp/bar/foo-test.bar*',
        naming.filepattern_for_dataset_split(
            dataset_name='foo',
            split=splits.Split.TEST,
            data_dir='/tmp/bar/',
            filetype_suffix='bar',
        ),
    )


def test_dataset_name_and_kwargs_from_name_str():
  assert naming._dataset_name_and_kwargs_from_name_str('ds1') == ('ds1', {})
  assert naming._dataset_name_and_kwargs_from_name_str('ds1:1.2.*') == (
      'ds1',
      {'version': '1.2.*'},
  )
  assert naming._dataset_name_and_kwargs_from_name_str('ds1/config1') == (
      'ds1',
      {'config': 'config1'},
  )
  assert naming._dataset_name_and_kwargs_from_name_str('ds1/config1:1.*.*') == (
      'ds1',
      {'config': 'config1', 'version': '1.*.*'},
  )
  assert naming._dataset_name_and_kwargs_from_name_str(
      'ds1/config1/arg1=val1,arg2=val2'
  ) == ('ds1', {'config': 'config1', 'arg1': 'val1', 'arg2': 'val2'})
  assert naming._dataset_name_and_kwargs_from_name_str(
      'ds1/config1:1.2.3/arg1=val1,arg2=val2'
  ) == (
      'ds1',
      {'config': 'config1', 'version': '1.2.3', 'arg1': 'val1', 'arg2': 'val2'},
  )
  assert naming._dataset_name_and_kwargs_from_name_str('ds1/arg1=val1') == (
      'ds1',
      {'arg1': 'val1'},
  )


def test_dataset_name():
  name = naming.DatasetName('ds1')
  assert name.name == 'ds1'
  assert name.namespace is None
  assert str(name) == 'ds1'
  assert repr(name) == "DatasetName('ds1')"

  name = naming.DatasetName('namespace123:ds1')
  assert name.name == 'ds1'
  assert name.namespace == 'namespace123'
  assert str(name) == 'namespace123:ds1'
  assert repr(name) == "DatasetName('namespace123:ds1')"

  name = naming.DatasetName(name='ds1', namespace='namespace123')
  assert name.name == 'ds1'
  assert name.namespace == 'namespace123'
  assert str(name) == 'namespace123:ds1'
  assert repr(name) == "DatasetName('namespace123:ds1')"

  name = naming.DatasetName(name='ds1')
  assert name.name == 'ds1'
  assert name.namespace is None
  assert str(name) == 'ds1'
  assert repr(name) == "DatasetName('ds1')"

  with pytest.raises(ValueError, match='Mixing args and kwargs'):
    name = naming.DatasetName('namespace123', name='abc')


@pytest.mark.parametrize(
    ['name', 'result'],
    [
        ('ds1', (naming.DatasetName('ds1'), {})),
        ('ds1:1.0.0', (naming.DatasetName('ds1'), {'version': '1.0.0'})),
        ('ns1:ds1', (naming.DatasetName('ns1:ds1'), {})),
        (
            'hugging_face:abc',
            (naming.DatasetName(namespace='hugging_face', name='abc'), {}),
        ),
        (
            'ns_1-b:ds1',
            (naming.DatasetName(namespace='ns_1-b', name='ds1'), {}),
        ),
        (
            'ns1:ds1:1.0.0',
            (naming.DatasetName('ns1:ds1'), {'version': '1.0.0'}),
        ),
        (
            'ns1:ds1/conf:1.0.0',
            (
                naming.DatasetName('ns1:ds1'),
                {
                    'version': '1.0.0',
                    'config': 'conf',
                },
            ),
        ),
        (
            'grand-vision:katr/128x128:1.0.0',
            (
                naming.DatasetName('grand-vision:katr'),
                {
                    'version': '1.0.0',
                    'config': '128x128',
                },
            ),
        ),
        (
            'huggingface:swiss_judgment_prediction/all+mt',
            (
                naming.DatasetName('huggingface:swiss_judgment_prediction'),
                {
                    'config': 'all+mt',
                },
            ),
        ),
    ],
)
def test_parse_builder_name_kwargs(name, result):
  assert naming.parse_builder_name_kwargs(name) == result


def test_parse_builder_name_kwargs_with_kwargs():
  parse = naming.parse_builder_name_kwargs

  assert parse('ds1', data_dir='/abc') == (
      naming.DatasetName('ds1'),
      {'data_dir': '/abc'},
  )

  with pytest.raises(TypeError, match='got multiple values for keyword arg'):
    parse('ds1:1.0.0', version='1.0.0')  # Version defined twice

  with pytest.raises(ValueError, match='Parsing builder name string .* failed'):
    parse('ds/config:ns:1.0.0')


def test_is_valid_dataset_name():
  assert naming.is_valid_dataset_name('dataset123_abc')
  assert not naming.is_valid_dataset_name('dataset-abc')
  assert not naming.is_valid_dataset_name('dataset.old')


def test_naming_sorted():
  assert sorted([
      naming.DatasetName('zzz:aaa'),
      naming.DatasetName('aaa:zzz'),
      naming.DatasetName('aaa:aaa'),
  ]) == [
      naming.DatasetName('aaa:aaa'),
      naming.DatasetName('aaa:zzz'),
      naming.DatasetName('zzz:aaa'),
  ]


def test_filename_info_with_small_shard_num():
  filename = 'mnist-test.tfrecord-00000-of-00001'
  assert naming.FilenameInfo.is_valid(filename)
  file_info = naming.FilenameInfo.from_str(filename)
  assert str(file_info) == filename
  assert file_info.dataset_name == 'mnist'
  assert file_info.split == 'test'
  assert file_info.filetype_suffix == 'tfrecord'
  assert file_info.shard_index == 0
  assert file_info.num_shards == 1


def test_filename_info_with_huge_shard_num():
  filenames = [
      'web_image_text-full.tfrecord-056234-of-104448',
      'web_image_text-full.tfrecord-56234-of-104448',  # backward compatible
  ]
  for filename in filenames:
    assert naming.FilenameInfo.is_valid(filename)
    file_info = naming.FilenameInfo.from_str(filename)
    assert str(file_info) == filenames[0]
    assert file_info.dataset_name == 'web_image_text'
    assert file_info.split == 'full'
    assert file_info.filetype_suffix == 'tfrecord'
    assert file_info.shard_index == 56234
    assert file_info.num_shards == 104448


def test_filename_info_with_custom_template():
  template = _FILENAME_TEMPLATE_CUSTOM_FULL
  filename = 'test.00042'
  file_info = naming.FilenameInfo.from_str(filename, filename_template=template)
  assert str(file_info) == filename
  assert file_info.dataset_name == 'mnist'
  assert file_info.split == 'test'
  assert file_info.filetype_suffix == 'tfrecord'
  assert file_info.shard_index == 42
  assert file_info.num_shards is None
  # This is not valid because the template is for the train split
  assert not naming.FilenameInfo.is_valid(filename, filename_template=template)
  assert naming.FilenameInfo.is_valid(
      filename, filename_template=template.replace(split='test')
  )


def test_filename_info_with_custom_template_missing_fields():
  template = _FILENAME_TEMPLATE_CUSTOM_NO_EXTRA
  filename = 'test.00042'
  assert naming.FilenameInfo.is_valid(filename, filename_template=template)
  file_info = naming.FilenameInfo.from_str(filename, filename_template=template)
  assert str(file_info) == filename
  assert file_info.dataset_name is None
  assert file_info.split == 'test'
  assert file_info.filetype_suffix is None
  assert file_info.shard_index == 42
  assert file_info.num_shards is None


@pytest.mark.parametrize(
    'filename',
    [
        'mnist-train.tfrecord-00000-of-00001',
        'mnist123-test.tfrecord-00032-of-01024',
        'mni23_st-test.riegeli-00032-of-01024',
    ],
)
def test_filename_info_valid(filename):
  assert naming.FilenameInfo.is_valid(filename)
  assert filename == str(naming.FilenameInfo.from_str(filename))


def test_filename_info_invalid():
  filename = 'mnist-train.tfrecord-000-of-001'  # Wrong shard number
  assert not naming.FilenameInfo.is_valid(filename)
  with pytest.raises(ValueError, match='Could not parse filename'):
    naming.FilenameInfo.from_str(filename)


def test_filename_info_with_path():
  filename_info = naming.FilenameInfo.from_str(
      '/path/mnist-train.tfrecord-00032-of-01024'
  )
  assert filename_info == naming.FilenameInfo(
      dataset_name='mnist',
      split='train',
      filetype_suffix='tfrecord',
      shard_index=32,
      num_shards=1024,
      filename_template=_FILENAME_TEMPLATE_DEFAULT.replace(data_dir='/path'),
  )


def test_filename_info_with_path_and_dash_in_split():
  filename_info = naming.FilenameInfo.from_str(
      '/path/c4-af-validation.tfrecord-00032-of-01024'
  )
  assert filename_info == naming.FilenameInfo(
      dataset_name='c4',
      split='af-validation',
      filetype_suffix='tfrecord',
      shard_index=32,
      num_shards=1024,
      filename_template=_FILENAME_TEMPLATE_DEFAULT.replace(data_dir='/path'),
  )


def test_filename_info_with_path_and_two_dashes_in_split():
  filename_info = naming.FilenameInfo.from_str(
      '/path/c4-bg-Latn-validation.tfrecord-00032-of-01024'
  )
  assert filename_info == naming.FilenameInfo(
      dataset_name='c4',
      split='bg-Latn-validation',
      filetype_suffix='tfrecord',
      shard_index=32,
      num_shards=1024,
      filename_template=_FILENAME_TEMPLATE_DEFAULT.replace(data_dir='/path'),
  )


def test_sharded_file_template_sharded_filepath():
  template = _FILENAME_TEMPLATE_MNIST_DEFAULT
  assert (
      os.fspath(template.sharded_filepath(shard_index=0, num_shards=1))
      == '/path/mnist-train.tfrecord-00000-of-00001'
  )
  assert (
      os.fspath(template.sharded_filepath(shard_index=0, num_shards=25))
      == '/path/mnist-train.tfrecord-00000-of-00025'
  )
  assert (
      os.fspath(template.sharded_filepath(shard_index=10, num_shards=25))
      == '/path/mnist-train.tfrecord-00010-of-00025'
  )


def test_sharded_file_template_empty_filetype_suffix():
  with pytest.raises(
      ValueError, match='Filetype suffix must be a non-empty string: .+'
  ):
    naming.ShardedFileTemplate(
        data_dir='/path', dataset_name='mnist', filetype_suffix=''
    )


def test_sharded_file_template_empty_split():
  with pytest.raises(ValueError, match='Split must be a non-empty string: .+'):
    naming.ShardedFileTemplate(
        data_dir='/path',
        dataset_name='mnist',
        filetype_suffix='tfrecord',
        split='',
    )


@pytest.mark.parametrize('split_name', [':', ',', '(())'])
def test_sharded_file_template_non_alphanumeric_split(split_name):
  match = (
      'Split name should contain at least one alphanumeric character. Given'
      f' split name: {split_name}'
  )
  with pytest.raises(
      ValueError,
      match=match,
  ):
    naming.ShardedFileTemplate(
        data_dir='/path',
        dataset_name='ds',
        filetype_suffix='',
        split=split_name,
    )


@pytest.mark.parametrize('split_name', ['train', '1', 'c', 'train_1', 'c:'])
def test_sharded_file_template_valid_split(split_name):
  naming.ShardedFileTemplate(
      data_dir='/path',
      dataset_name='mnist',
      filetype_suffix='tfrecord',
      split=split_name,
  )


def test_sharded_file_template_shard_index():
  builder_dir = epath.Path('/my/path')
  template = naming.ShardedFileTemplate(
      template='data/mnist-train.tfrecord-{SHARD_INDEX}', data_dir=builder_dir
  )
  assert (
      os.fspath(template.sharded_filepath(shard_index=12, num_shards=100))
      == '/my/path/data/mnist-train.tfrecord-00012'
  )
  assert (
      os.fspath(template.sharded_filepaths_pattern())
      == '/my/path/data/mnist-train.tfrecord*'
  )
  assert (
      os.fspath(template.sharded_filepaths_pattern(num_shards=100))
      == '/my/path/data/mnist-train.tfrecord@100'
  )


def test_sharded_file_template_sharded_filepath_shard_x_of_y():
  builder_dir = epath.Path('/my/path')
  template_explicit = naming.ShardedFileTemplate(
      template='data/mnist-train.tfrecord-{SHARD_INDEX}-of-{NUM_SHARDS}',
      data_dir=builder_dir,
  )
  assert (
      os.fspath(
          template_explicit.sharded_filepath(shard_index=12, num_shards=100)
      )
      == '/my/path/data/mnist-train.tfrecord-00012-of-00100'
  )

  template = naming.ShardedFileTemplate(
      template='data/mnist-train.tfrecord-{SHARD_X_OF_Y}', data_dir=builder_dir
  )
  assert (
      os.fspath(template.sharded_filepath(shard_index=12, num_shards=100))
      == '/my/path/data/mnist-train.tfrecord-00012-of-00100'
  )


def test_sharded_file_template_sharded_filepath_shard_x_of_y_more_digits():
  builder_dir = epath.Path('/my/path')
  template = naming.ShardedFileTemplate(
      template='data/{DATASET}-{SPLIT}.{FILEFORMAT}-{SHARD_X_OF_Y}',
      data_dir=builder_dir,
      dataset_name='mnist',
      filetype_suffix='tfrecord',
      split='train',
  )
  assert (
      os.fspath(template.sharded_filepath(shard_index=12, num_shards=1234567))
      == '/my/path/data/mnist-train.tfrecord-0000012-of-1234567'
  )


def test_sharded_file_template_no_template():
  builder_dir = epath.Path('/my/path')
  template = naming.ShardedFileTemplate(
      data_dir=builder_dir,
      dataset_name='imagenet',
      filetype_suffix='riegeli',
      split='test',
  )
  assert (
      os.fspath(template.sharded_filepath(shard_index=12, num_shards=100))
      == '/my/path/imagenet-test.riegeli-00012-of-00100'
  )


def test_sharded_file_template_no_template_incomplete():
  builder_dir = epath.Path('/my/path')
  template_without_split = naming.ShardedFileTemplate(
      data_dir=builder_dir, dataset_name='imagenet', filetype_suffix='riegeli'
  )
  with pytest.raises(KeyError):
    template_without_split.sharded_filepath(shard_index=12, num_shards=100)


def test_sharded_file_template_template_and_properties():
  builder_dir = epath.Path('/my/path')
  template = naming.ShardedFileTemplate(
      template='data/mnist-{SPLIT}.{FILEFORMAT}-{SHARD_INDEX}',
      data_dir=builder_dir,
      # dataset_name is ignored because the template doesn't have {DATASET}
      dataset_name='imagenet',
      filetype_suffix='riegeli',
      split='test',
  )
  assert (
      os.fspath(template.sharded_filepath(shard_index=12, num_shards=100))
      == '/my/path/data/mnist-test.riegeli-00012'
  )


def test_replace_shard_suffix():
  assert (
      naming._replace_shard_suffix(
          filepath='/path/file-00001-of-01234', replacement='repl'
      )
      == '/path/filerepl'
  )
  assert (
      naming._replace_shard_suffix(
          filepath='/path/file00001-of-01234', replacement='repl'
      )
      == '/path/filerepl'
  )
  assert (
      naming._replace_shard_suffix(
          filepath='/path/file-00001', replacement='repl'
      )
      == '/path/filerepl'
  )


def test_replace_shard_suffix_folder_similar_to_shard():
  assert (
      naming._replace_shard_suffix(
          filepath='/path/i-look-like-a-shard-000001/file-00001-of-01234',
          replacement='repl',
      )
      == '/path/i-look-like-a-shard-000001/filerepl'
  )


def test_replace_shard_suffix_no_suffix_found():
  with pytest.raises(RuntimeError, match='Should do 1 shard suffix'):
    naming._replace_shard_suffix(filepath='/path/a/b', replacement='')


def test_filename_template_to_regex():
  assert (
      naming._filename_template_to_regex('{DATASET}')
      == r'(?P<dataset_name>[a-zA-Z][\w]*)'
  )
  assert naming._filename_template_to_regex(
      naming.DEFAULT_FILENAME_TEMPLATE
  ) == (
      r'(?P<dataset_name>[a-zA-Z][\w]*)'
      r'-(?P<split>(\w|-)+)'
      r'\.(?P<filetype_suffix>\w+)'
      r'-(?P<shard_index>\d{5,})-of-(?P<num_shards>\d{5,})'
  )


def test_filename_template_to_regex_unknown_variables():
  with pytest.raises(ValueError, match='Regex still contains variables.+'):
    naming._filename_template_to_regex('{XYZ}')


@pytest.mark.parametrize(
    ['name', 'result'],
    [
        ('mnist-train.tfrecord-00000-of-00001', True),
        ('xyz-train.tfrecord-00000-of-00001', False),
        ('mnist-xyz.tfrecord-00000-of-00001', False),
        ('mnist-train.xyz-00000-of-00001', False),
        ('mnist-train.tfrecord-a-of-b', False),
    ],
)
def test_sharded_file_template_is_valid_default_template(name, result):
  template = _FILENAME_TEMPLATE_MNIST_DEFAULT
  assert template.is_valid(name) == result


@pytest.mark.parametrize(
    ['name', 'result'],
    [
        ('train.1', False),
        ('train.00001', True),
        ('train-00001', False),
        ('train.tfrecord.00001', False),
        ('mnist-train.tfrecord-00000-of-00001', False),
    ],
)
def test_sharded_file_template_is_valid_custom_template(name, result):
  template = _FILENAME_TEMPLATE_CUSTOM_FULL
  assert template.is_valid(name) == result


@pytest.mark.parametrize(
    ['name', 'result'],
    [
        (
            'mnist-train.tfrecord-00000-of-00001',
            naming.FilenameInfo(
                dataset_name='mnist',
                split='train',
                filetype_suffix='tfrecord',
                shard_index=0,
                num_shards=1,
                filename_template=_FILENAME_TEMPLATE_MNIST_DEFAULT,
            ),
        ),
        (
            'mnist-train.tfrecord-00123-of-00456',
            naming.FilenameInfo(
                dataset_name='mnist',
                split='train',
                filetype_suffix='tfrecord',
                shard_index=123,
                num_shards=456,
                filename_template=_FILENAME_TEMPLATE_MNIST_DEFAULT,
            ),
        ),
        ('train1', None),
    ],
)
def test_sharded_file_template_parse_filename_info(name, result):
  template = _FILENAME_TEMPLATE_MNIST_DEFAULT
  assert template.parse_filename_info(name) == result


@pytest.mark.parametrize(
    ['name', 'result'],
    [
        (
            'train.00123',
            naming.FilenameInfo(
                dataset_name=None,
                split='train',
                filetype_suffix=None,
                shard_index=123,
                num_shards=None,
                filename_template=_FILENAME_TEMPLATE_CUSTOM_NO_EXTRA,
            ),
        ),
        ('mnist-train.tfrecord-00000-of-00001', None),
        ('train1', None),
    ],
)
def test_sharded_file_template_parse_filename_info_custom_template(
    name, result
):
  template = _FILENAME_TEMPLATE_CUSTOM_NO_EXTRA
  assert template.parse_filename_info(name) == result


@pytest.mark.parametrize(
    ['name', 'result'],
    [
        (
            'train.00123',
            naming.FilenameInfo(
                dataset_name='mnist',
                split='train',
                filetype_suffix='tfrecord',
                shard_index=123,
                num_shards=None,
                filename_template=_FILENAME_TEMPLATE_CUSTOM_FULL,
            ),
        ),
        (
            'test.00042',
            naming.FilenameInfo(
                dataset_name='mnist',
                split='test',
                filetype_suffix='tfrecord',
                shard_index=42,
                num_shards=None,
                filename_template=_FILENAME_TEMPLATE_CUSTOM_FULL,
            ),
        ),
        ('mnist-train.tfrecord-00000-of-00001', None),
        ('train1', None),
    ],
)
def test_sharded_file_template_parse_filename_info_custom_template_add_missing(
    name, result
):
  template = _FILENAME_TEMPLATE_CUSTOM_FULL
  assert template.parse_filename_info(name) == result


@pytest.mark.parametrize(
    (
        'tfds_name',
        'namespace',
        'split_mapping',
        'data_dir',
        'ds_name',
        'version',
        'config',
    ),
    [
        # Dataset with a config and a version.
        ('ds/config:1.2.3', None, None, None, 'ds', '1.2.3', 'config'),
        # Dataset with a config and a version and a data_dir.
        ('ds/config:1.2.3', None, None, '/a/b', 'ds', '1.2.3', 'config'),
        # Test having a split mapping.
        ('ds/config:1.2.3', None, {'x': 'y'}, None, 'ds', '1.2.3', 'config'),
        # Dataset without a config but with a version.
        ('ds:1.2.3', None, None, None, 'ds', '1.2.3', None),
        # Dataset with a config but without a version.
        ('ds/config', None, None, None, 'ds', None, 'config'),
        # Dataset without a config and a version.
        ('ds', None, None, None, 'ds', None, None),
        # Dataset with a namespace.
        ('ns:ds/config:1.2.3', 'ns', None, '/a/b', 'ds', '1.2.3', 'config'),
    ],
)
def test_dataset_reference_from_tfds_name(
    tfds_name, namespace, split_mapping, data_dir, ds_name, version, config
):
  actual = naming.DatasetReference.from_tfds_name(
      tfds_name=tfds_name, split_mapping=split_mapping, data_dir=data_dir
  )
  assert actual == naming.DatasetReference(
      dataset_name=ds_name,
      namespace=namespace,
      version=version,
      config=config,
      split_mapping=split_mapping,
      data_dir=data_dir,
  )


@pytest.mark.parametrize(
    ('ds_name', 'namespace', 'version', 'config', 'tfds_name'),
    [
        ('ds', 'ns', '1.2.3', 'config', 'ns:ds/config:1.2.3'),
        ('ds', None, '1.2.3', 'config', 'ds/config:1.2.3'),
        ('ds', None, '1.2.3', None, 'ds:1.2.3'),
        ('ds', None, None, None, 'ds'),
    ],
)
def test_dataset_reference_tfds_name(
    ds_name, namespace, version, config, tfds_name
):
  reference = naming.DatasetReference(
      dataset_name=ds_name, namespace=namespace, version=version, config=config
  )
  assert reference.tfds_name() == tfds_name


def test_dataset_reference_tfds_name_without_version():
  reference = naming.DatasetReference(
      dataset_name='ds', version='1.2.3', config='config'
  )
  assert reference.tfds_name(include_version=False) == 'ds/config'


@pytest.mark.parametrize(
    ('tfds_name', 'data_dir', 'dataset_dir'),
    [
        ('ns:ds/config:1.2.3', '/a/b', '/a/b/ds/config/1.2.3'),
        ('ds/config:1.2.3', '/a/b', '/a/b/ds/config/1.2.3'),
        ('ds:1.2.3', '/a/b', '/a/b/ds/1.2.3'),
    ],
)
def test_dataset_reference_dataset_dir(tfds_name, data_dir, dataset_dir):
  reference = naming.DatasetReference.from_tfds_name(
      tfds_name=tfds_name, data_dir=data_dir
  )
  assert os.fspath(reference.dataset_dir()) == dataset_dir

  # Also test passing the data_dir to `.dataset_dir`
  reference = naming.DatasetReference.from_tfds_name(
      tfds_name=tfds_name, data_dir='/something/else'
  )
  assert os.fspath(reference.dataset_dir(data_dir=data_dir)) == dataset_dir


def test_dataset_reference_get_split():
  reference = naming.DatasetReference.from_tfds_name(
      'ds/config:1.2.3', split_mapping={'x': 'y'}
  )
  assert reference.get_split('x') == 'y'
  assert reference.get_split('y') == 'y'
  assert reference.get_split('z') == 'z'


def test_references_for():
  expected = {
      'one': naming.DatasetReference(
          dataset_name='ds1', version='1.2.3', config='config'
      ),
      'two': naming.DatasetReference(dataset_name='ds2', version='1.0.0'),
  }
  assert (
      naming.references_for({'one': 'ds1/config:1.2.3', 'two': 'ds2:1.0.0'})
      == expected
  )


def test_reference_for():
  expected = naming.DatasetReference(
      dataset_name='ds', version='1.2.3', config='config'
  )
  assert naming.reference_for('ds/config:1.2.3') == expected



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""genomics_ood dataset."""

import csv
import os

from etils import epath
import tensorflow_datasets.public_api as tfds

_CITATION = """
@inproceedings{ren2019likelihood,
  title={Likelihood ratios for out-of-distribution detection},
  author={Ren, Jie and
  Liu, Peter J and
  Fertig, Emily and
  Snoek, Jasper and
  Poplin, Ryan and
  Depristo, Mark and
  Dillon, Joshua and
  Lakshminarayanan, Balaji},
  booktitle={Advances in Neural Information Processing Systems},
  pages={14707--14718},
  year={2019}
}
"""

_DESCRIPTION = """
Bacteria identification based on genomic sequences holds the promise of early
detection of diseases, but requires a model that can output low confidence
predictions on out-of-distribution (OOD) genomic sequences from new bacteria
that were not present in the training data.

We introduce a genomics dataset for OOD detection that allows other researchers
to benchmark progress on this important problem. New bacterial classes are
gradually discovered over the years. Grouping classes by years is a natural way
to mimic the in-distribution and OOD examples.

The dataset contains genomic sequences sampled from 10 bacteria classes that
were discovered before the year 2011 as in-distribution classes, 60 bacteria
classes discovered between 2011-2016 as OOD for validation, and another 60
different bacteria classes discovered after 2016 as OOD for test, in total 130
bacteria classes. Note that training, validation, and test data are provided for
the in-distribution classes, and validation and test data are proviede for OOD
classes. By its nature, OOD data is not available at the training time.

The genomic sequence is 250 long, composed by characters of {A, C, G, T}. The
sample size of each class is 100,000 in the training and 10,000 for the
validation and test sets.

For each example, the features include:
  seq: the input DNA sequence composed by {A, C, G, T}.
  label: the name of the bacteria class.
  seq_info: the source of the DNA sequence, i.e., the genome name, NCBI
  accession number, and the position where it was sampled from.
  domain: if the bacteria is in-distribution (in), or OOD (ood)

The details of the dataset can be found in the paper supplemental.
"""

_DATA_URL = tfds.core.gcs_path('downloads/genomics_ood/genomics_ood.zip')

_LABELS_IN = [
    'Bacillus',
    'Burkholderia',
    'Clostridium',
    'Escherichia',
    'Mycobacterium',
    'Pseudomonas',
    'Salmonella',
    'Staphylococcus',
    'Streptococcus',
    'Yersinia',
]
_LABELS_OOD_VAL = [
    'Actinoplanes',
    'Advenella',
    'Alicycliphilus',
    'Altererythrobacter',
    'Anabaena',
    'Archangium',
    'Bibersteinia',
    'Blastochloris',
    'Calothrix',
    'Carnobacterium',
    'Cedecea',
    'Cellulophaga',
    'Chondromyces',
    'Chryseobacterium',
    'Collimonas',
    'Corallococcus',
    'Cyclobacterium',
    'Dehalobacter',
    'Desulfosporosinus',
    'Devosia',
    'Dyella',
    'Elizabethkingia',
    'Glaciecola',
    'Granulicella',
    'Haliscomenobacter',
    'Hymenobacter',
    'Kibdelosporangium',
    'Kutzneria',
    'Labilithrix',
    'Leptolyngbya',
    'Leptospirillum',
    'Lysobacter',
    'Mannheimia',
    'Massilia',
    'Methanobacterium',
    'Microbacterium',
    'Myroides',
    'Neorhizobium',
    'Niastella',
    'Oblitimonas',
    'Octadecabacter',
    'Oscillatoria',
    'Pandoraea',
    'Pelosinus',
    'Phaeobacter',
    'Piscirickettsia',
    'Planococcus',
    'Pseudonocardia',
    'Pseudoxanthomonas',
    'Rahnella',
    'Raoultella',
    'Rufibacter',
    'Saccharothrix',
    'Sandaracinus',
    'Singulisphaera',
    'Sphaerochaeta',
    'Sphingobacterium',
    'Spiroplasma',
    'Tannerella',
    'Terriglobus',
]
_LABELS_OOD_TEST = [
    'Actinoalloteichus',
    'Aeromicrobium',
    'Agromyces',
    'Aminobacter',
    'Aneurinibacillus',
    'Blastomonas',
    'Blautia',
    'Bosea',
    'Brevibacterium',
    'Cellulosimicrobium',
    'Chryseolinea',
    'Cryobacterium',
    'Cystobacter',
    'Dietzia',
    'Ensifer',
    'Faecalibacterium',
    'Fictibacillus',
    'Filimonas',
    'Flammeovirga',
    'Fuerstia',
    'Gemmata',
    'Granulosicoccus',
    'Halioglobus',
    'Hydrogenophaga',
    'Labrenzia',
    'Leclercia',
    'Lelliottia',
    'Lentzea',
    'Luteitalea',
    'Melittangium',
    'Microbulbifer',
    'Microvirga',
    'Minicystis',
    'Moorea',
    'Mucilaginibacter',
    'Natronolimnobius',
    'Nitratireductor',
    'Nitrospirillum',
    'Nonomuraea',
    'Olleya',
    'Paludisphaera',
    'Pannonibacter',
    'Petrimonas',
    'Planctomyces',
    'Plantactinospora',
    'Plesiomonas',
    'Porphyrobacter',
    'Rhizobacter',
    'Rhodoplanes',
    'Roseomonas',
    'Roseovarius',
    'Salinimonas',
    'Shinella',
    'Sphingorhabdus',
    'Sporosarcina',
    'Sulfitobacter',
    'Tatumella',
    'Tessaracoccus',
    'Thiodictyon',
    'Tumebacillus',
]
_LABELS_ALL = _LABELS_IN + _LABELS_OOD_VAL + _LABELS_OOD_TEST


class GenomicsOod(tfds.core.GeneratorBasedBuilder):
  """Genomic sequence dataset for out-of-distribution (OOD) detection."""

  VERSION = tfds.core.Version('0.0.1')

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        # This is the description that will appear on the datasets page.
        description=_DESCRIPTION,
        # tfds.features.FeatureConnectors
        features=tfds.features.FeaturesDict({
            'seq': tfds.features.Text(),
            'label': tfds.features.ClassLabel(names=_LABELS_ALL),
            'seq_info': tfds.features.Text(),
            'domain': tfds.features.Text(),
        }),
        # If there's a common (input, target) tuple from the features,
        # specify them here. They'll be used if as_supervised=True in
        # builder.as_dataset.
        supervised_keys=('seq', 'label'),
        # Homepage of the dataset for documentation
        homepage='https://github.com/google-research/google-research/tree/master/genomics_ood',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    data_path = dl_manager.extract(_DATA_URL)
    return [
        tfds.core.SplitGenerator(
            name='train',  # name=tfds.Split.TRAIN,
            gen_kwargs={
                'filename': os.path.join(data_path, 'before_2011_in_tr.txt')
            },
        ),
        tfds.core.SplitGenerator(
            name='validation',
            gen_kwargs={
                'filename': os.path.join(
                    data_path, 'between_2011-2016_in_val.txt'
                )
            },
        ),
        tfds.core.SplitGenerator(
            name='test',
            gen_kwargs={
                'filename': os.path.join(data_path, 'after_2016_in_test.txt')
            },
        ),
        tfds.core.SplitGenerator(
            name='validation_ood',
            gen_kwargs={
                'filename': os.path.join(
                    data_path, 'between_2011-2016_ood_val.txt'
                )
            },
        ),
        tfds.core.SplitGenerator(
            name='test_ood',
            gen_kwargs={
                'filename': os.path.join(data_path, 'after_2016_ood_test.txt')
            },
        ),
    ]

  def _generate_examples(self, filename):
    """Yields examples."""
    with epath.Path(filename).open() as f:
      reader = csv.DictReader(f, delimiter='\t')
      for row_id, row in enumerate(reader):
        example = {}
        example['seq'] = row['seq']
        example['label'] = row['label']
        example['seq_info'] = row['seq_info']
        example['domain'] = row['domain']

        yield row_id, example



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for qa_utils."""

import os

from tensorflow_datasets import testing
from tensorflow_datasets.question_answering import qa_utils


class QAUtilsTest(testing.TestCase):

  def test_generate_squadlike_examples(self):
    filepath = os.path.join(
        testing.test_utils.fake_examples_dir(), 'xquad', 'translate-test.json'
    )
    examples = qa_utils.generate_squadlike_examples(filepath)

    self.assertEqual(
        list(examples),
        [
            (
                '1',
                {
                    'id': '1',
                    'title': 'Zurich_Switzerland',
                    'context': (
                        'Zurich is the largest city in Switzerland with over'
                        ' 400000 inhabitants. In spite of this, it is not the'
                        ' capital of Switzerland, which is located in Bern aka'
                        ' Bernie.'
                    ),
                    'question': 'What is the capital of Switzerland?',
                    'answers': {
                        'answer_start': [1, 20, 29],
                        'text': ['Zurich', 'Bern', 'Bernie'],
                    },
                },
            ),
            (
                '2',
                {
                    'id': '2',
                    'title': 'Zurich_Switzerland',
                    'context': (
                        'Switzerland is the country in Euriope with 26 cantons.'
                        ' Zurich canton has the largest population of 1.5'
                        ' million.'
                    ),
                    'question': 'How many cantons does Switzerland have?',
                    'answers': {'answer_start': [8], 'text': ['26']},
                },
            ),
            (
                '3',
                {
                    'id': '3',
                    'title': 'Paris_France',
                    'context': (
                        '  Paris is the largest city in France with over 2'
                        ' million inhabitants. It is the capital of France.'
                    ),
                    'question': 'What is the capital of France?',
                    'answers': {
                        'answer_start': [3, 9],
                        'text': ['Paris', 'France'],
                    },
                },
            ),
        ],
    )


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils functions."""

import enum


class GenerateMode(enum.Enum):
  """`Enum` for how to treat pre-existing downloads and data.

  The default mode is `REUSE_DATASET_IF_EXISTS`, which will reuse both
  raw downloads and the prepared dataset if they exist.

  The generations modes:

  |                                    | Downloads | Dataset | Metadata |
  | -----------------------------------|-----------|---------|----------|
  | `REUSE_DATASET_IF_EXISTS` (default)| Reuse     | Reuse   | Reuse    |
  | `UPDATE_DATASET_INFO`              | Reuse     | Reuse   | Fresh    |
  | `REUSE_CACHE_IF_EXISTS`            | Reuse     | Fresh   | Fresh    |
  | `FORCE_REDOWNLOAD`                 | Fresh     | Fresh   | Fresh    |

  UPDATE_DATASET_INFO only regenerates DatasetInfo metadata which is directly
  coming from the Builder metadata, and not directly used to prepare the data
  or computed from the downloaded or prepared data.
  This means that `description`, `config_tags`, etc. will be updated, but
  `download_size`, `schema`, `splits`, `disable_shuffling`, `file_format` will
  not be updated.
  UPDATE_DATASET_INFO will fail if the data has never been prepared.
  """

  REUSE_DATASET_IF_EXISTS = 'reuse_dataset_if_exists'
  UPDATE_DATASET_INFO = 'update_dataset_info'
  REUSE_CACHE_IF_EXISTS = 'reuse_cache_if_exists'
  FORCE_REDOWNLOAD = 'force_redownload'


class ComputeStatsMode(enum.Enum):
  """Mode to decide if dynamic dataset info fields should be computed or not.

  Mode can be:

  * AUTO: Compute the DatasetInfo dynamic fields only if they haven't been
    restored from GCS.
  * FORCE: Always recompute DatasetInfo dynamic  fields, even if they are
    already present
  * SKIP: Ignore the dataset dynamic field computation (whether they already
    exist or not)
  """

  AUTO = 'auto'
  FORCE = 'force'
  SKIP = 'skip'



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""d4rl_adroit_pen dataset."""
from typing import Any

from tensorflow_datasets.d4rl import dataset_builder
import tensorflow_datasets.public_api as tfds


class D4rlAdroitPen(dataset_builder.D4RLDatasetBuilder):
  """DatasetBuilder for d4rl_adroit_pen dataset."""

  VERSION = tfds.core.Version('1.1.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
      '1.1.0': 'Added is_last.',
  }

  # pytype: disable=wrong-keyword-args
  # pylint: disable=protected-access
  # Pen uses a different policy size in expert-v1
  BUILDER_CONFIGS = dataset_builder.ADROIT_BUILDER_CONFIGS[:-1] + [
      dataset_builder.BuilderConfig(
          name='v1-expert',
          dataset_dir='hand_dapg_v1',
          env='adroit',
          file_suffix='-expert-v1',
          step_metadata_keys=frozenset([
              dataset_builder._QPOS,
              dataset_builder._QVEL,
              dataset_builder._ADROIT_BODY_POS,
              dataset_builder._ACTION_MEAN,
              dataset_builder._ACTION_LOG_STD,
          ]),
          episode_metadata_keys=frozenset([dataset_builder._ALGORITHM]),
          has_policy_metadata=True,
          has_policy_last_fc_log_std=True,
          policy_size=64,  # pytype: disable=wrong-arg-types  # gen-stub-imports
      ),
  ]

  # pylint: enable=protected-access
  # pytype: enable=wrong-keyword-args

  def __init__(self, **kwargs: Any):
    config = dataset_builder.DatasetConfig(
        name='pen', obs_len=45, action_len=24, qpos_len=30, qvel_len=30
    )
    super().__init__(ds_config=config, **kwargs)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Copy the info files from placer to GCS bucket.
"""

import os

from absl import app
from absl import flags
from absl import logging
import tensorflow_datasets as tfds
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

flags.DEFINE_boolean('dry_run', True, 'If True, just print, do nothing.')
flags.DEFINE_boolean('overwrite', False, 'If True, overwrites the data.')
flags.DEFINE_string(
    'from_directory',
    tfds.core.constants.DATA_DIR,
    'Where to get the info files from (datasets/ dir on placer).',
)
flags.DEFINE_string(
    'to_directory', None, 'Path where dataset info files will be copied.'
)

FLAGS = flags.FLAGS


def _copy_metadata(from_dir, to_dir):
  """Copy the info files from within `from_dir` to `to_dir`."""
  if not FLAGS.dry_run:
    tf.io.gfile.makedirs(to_dir)
  for fname in tfds.core.utils.list_info_files(from_dir):
    from_path = os.path.join(from_dir, fname)
    to_path = os.path.join(to_dir, fname)
    logging.info('cp %s %s', from_path, to_path)
    if not FLAGS.dry_run:
      tf.io.gfile.copy(from_path, to_path, overwrite=True)


def copy(from_dir: tfds.typing.PathLike, to_dir: tfds.typing.PathLike) -> None:
  """Copy the info files from within `from_dir` to `to_dir`."""
  for full_name in tfds.core.load.list_full_names():
    from_full_name_dir = os.path.join(from_dir, full_name)
    to_full_name_dir = os.path.join(to_dir, full_name)

    # Skip if the dataset isn't generated or that metadata are already copied
    if not tf.io.gfile.exists(from_full_name_dir):
      logging.info('Skipping %s (not found)', from_full_name_dir)
      continue
    if tf.io.gfile.exists(to_full_name_dir) and not FLAGS.overwrite:
      logging.info('Skipping %s (already exists)', to_full_name_dir)
      continue

    _copy_metadata(from_dir=from_full_name_dir, to_dir=to_full_name_dir)


def main(_):
  copy(
      FLAGS.from_directory,
      FLAGS.to_directory or tfds.core.gcs_path('dataset_info'),
  )


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Texture tiles from colorectal cancer histology."""

import os

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_URL = "https://zenodo.org/record/53169#.XGZemKwzbmG"
_TILES_DL_URL = "https://zenodo.org/record/53169/files/Kather_texture_2016_image_tiles_5000.zip"
_LARGE_DL_URL = "https://zenodo.org/record/53169/files/Kather_texture_2016_larger_images_10.zip"

_TILES_SUBDIR = "Kather_texture_2016_image_tiles_5000"
_LARGE_SUBDIR = "Kather_texture_2016_larger_images_10"

_CLASS_NAMES = (
    "tumor",
    "stroma",
    "complex",
    "lympho",
    "debris",
    "mucosa",
    "adipose",
    "empty",
)
_TILES_SIZE = 150
_LARGE_SIZE = 5000

_CITATION = """\
@article{kather2016multi,
  title={Multi-class texture analysis in colorectal cancer histology},
  author={Kather, Jakob Nikolas and Weis, Cleo-Aron and Bianconi, Francesco and Melchers, Susanne M and Schad, Lothar R and Gaiser, Timo and Marx, Alexander and Z{\"o}llner, Frank Gerrit},
  journal={Scientific reports},
  volume={6},
  pages={27988},
  year={2016},
  publisher={Nature Publishing Group}
}
"""


def _class_subdir(class_index, class_name):
  return "%02d_%s" % (class_index + 1, class_name.upper())


def _load_tif(path):
  with tf.io.gfile.GFile(path, "rb") as fp:
    image = tfds.core.lazy_imports.PIL_Image.open(fp)
  return np.array(image)


class ColorectalHistology(tfds.core.GeneratorBasedBuilder):
  """Biological 8-class classification problem."""

  VERSION = tfds.core.Version("2.0.0")
  RELEASE_NOTES = {
      "2.0.0": "New split API (https://tensorflow.org/datasets/splits)",
  }

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=(
            "Classification of textures in colorectal cancer histology. "
            "Each example is a 150 x 150 x 3 RGB image of one of 8 classes."
        ),
        features=tfds.features.FeaturesDict({
            "image": tfds.features.Image(shape=(_TILES_SIZE,) * 2 + (3,)),
            "label": tfds.features.ClassLabel(names=_CLASS_NAMES),
            "filename": tfds.features.Text(),
        }),
        homepage=_URL,
        citation=_CITATION,
        supervised_keys=("image", "label"),
    )

  def _split_generators(self, dl_manager):
    folder = dl_manager.download_and_extract(_TILES_DL_URL)
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs=dict(root_dir=folder),
        ),
    ]

  def _generate_examples(self, root_dir):
    root_dir = os.path.join(root_dir, _TILES_SUBDIR)
    for i, class_name in enumerate(_CLASS_NAMES):
      class_dir = os.path.join(root_dir, _class_subdir(i, class_name))
      fns = tf.io.gfile.listdir(class_dir)

      for fn in sorted(fns):
        image = _load_tif(os.path.join(class_dir, fn))
        record = {
            "image": image,
            "label": class_name,
            "filename": fn,
        }
        yield "%s/%s" % (class_name, fn), record


class ColorectalHistologyLarge(tfds.core.GeneratorBasedBuilder):
  """10 Large 5000 x 5000 colorectal histology images without labels."""

  VERSION = tfds.core.Version("2.0.0")
  RELEASE_NOTES = {
      "2.0.0": "New split API (https://tensorflow.org/datasets/splits)",
  }

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=(
            "10 large 5000 x 5000 textured colorectal cancer histology images"
        ),
        features=tfds.features.FeaturesDict({
            "image": tfds.features.Image(shape=(_LARGE_SIZE,) * 2 + (3,)),
            "filename": tfds.features.Text(),
        }),
        homepage=_URL,
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    folder = dl_manager.download_and_extract(_LARGE_DL_URL)
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST, gen_kwargs=dict(folder=folder)
        )
    ]

  def _generate_examples(self, folder):
    folder = os.path.join(folder, _LARGE_SUBDIR)
    for fn in tf.io.gfile.listdir(folder):
      image = _load_tif(os.path.join(folder, fn))
      record = dict(image=image, filename=fn)
      yield fn, record



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Show example util."""

from __future__ import annotations

from collections.abc import Iterable
import typing
from typing import Any, Union

from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core import splits
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
from tensorflow_datasets.core.visualization import graph_visualizer
from tensorflow_datasets.core.visualization import image_visualizer

from tensorflow_metadata.proto.v0 import statistics_pb2

if typing.TYPE_CHECKING:
  _Dataset = Union[
      tf.data.Dataset,
      Iterable,
  ]
else:
  _Dataset = Any

_ALL_VISUALIZERS = [
    image_visualizer.ImageGridVisualizer(),
    graph_visualizer.GraphVisualizer(),
]

_DEFAULT_NUM_COLS = 3
_DEFAULT_NUM_ROWS = 3


def _to_tf_dataset(
    ds: _Dataset,
    min_length: int,
    is_batched: bool = False,
) -> tf.data.Dataset:
  """Converts any iterable to a small tf.data.Dataset to use visualizations.

  Warning: this util function is not optimized, so it should only be used for a
  small number of records (i.e., small `min_length`).

  Args:
    ds: Any dataset as an iterable.
    min_length: The minimum number of examples to generate.
    is_batched: Whether the data is batched.

  Returns:
    the tf.data.Dataset of cardinality at least `min_length`.
  """
  if isinstance(ds, tf.data.Dataset) and not isinstance(ds, Iterable):
    if is_batched:
      return ds.unbatch()
    else:
      return ds
  tf_dataset = None
  if is_batched:
    from_tensor = tf.data.Dataset.from_tensor_slices
  else:
    from_tensor = tf.data.Dataset.from_tensors
  for record in ds:
    if tf_dataset is None:
      tf_dataset = from_tensor(record)
    else:
      tf_dataset = tf_dataset.concatenate(from_tensor(record))
    # Terminate if `tf_dataset` reached at least the expected `min_length`.
    if tf_dataset.cardinality().numpy() >= min_length:
      break
  if tf_dataset is None:
    raise ValueError(
        'Empty dataset, could not generate a valid tf.data.Dataset.'
    )
  return tf_dataset


def show_examples(
    ds: _Dataset,
    ds_info: dataset_info.DatasetInfo,
    is_batched: bool = False,
    **options_kwargs: Any,
):
  """Visualize images (and labels) from an image classification dataset.

  This function is for interactive use (Colab, Jupyter). It displays and return
  a plot of (rows*columns) images from a tf.data.Dataset.

  Usage:
  ```python
  ds, ds_info = tfds.load('cifar10', split='train', with_info=True)
  fig = tfds.show_examples(ds, ds_info)
  ```

  Args:
    ds: `tf.data.Dataset`. The tf.data.Dataset object to visualize. Examples
      should not be batched. Examples will be consumed in order until (rows *
      cols) are read or the dataset is consumed.
    ds_info: The dataset info object to which extract the label and features
      info. Available either through `tfds.load('mnist', with_info=True)` or
      `tfds.builder('mnist').info`
    is_batched: Whether the data is batched.
    **options_kwargs: Additional display options, specific to the dataset type
      to visualize. Are forwarded to `tfds.visualization.Visualizer.show`. See
      the `tfds.visualization` for a list of available visualizers.

  Returns:
    fig: The `matplotlib.Figure` object
  """
  rows = options_kwargs.pop('rows', _DEFAULT_NUM_ROWS)
  cols = options_kwargs.pop('cols', _DEFAULT_NUM_COLS)
  ds = _to_tf_dataset(ds, rows * cols, is_batched=is_batched)
  if not isinstance(ds_info, dataset_info.DatasetInfo):  # Arguments inverted
    # `absl.logging` does not appear on Colab by default, so uses print instead.
    print(
        'WARNING: For consistency with `tfds.load`, the `tfds.show_examples` '
        'signature has been modified from (info, ds) to (ds, info).\n'
        'The old signature is deprecated and will be removed. '
        'Please change your call to `tfds.show_examples(ds, info)`'
    )
    ds, ds_info = ds_info, ds

  # Pack `as_supervised=True` datasets
  ds = dataset_info.pack_as_supervised_ds(ds, ds_info)

  for visualizer in _ALL_VISUALIZERS:
    if visualizer.match(ds_info):
      return visualizer.show(
          ds, ds_info, **options_kwargs, rows=rows, cols=cols
      )

  raise ValueError(
      'Visualisation not supported for dataset `{}`'.format(ds_info.name)
  )


def show_statistics(
    ds_info: dataset_info.DatasetInfo,
    split: splits.Split = splits.Split.TRAIN,
    disable_logging: bool = True,
) -> None:
  """Display the datasets statistics on a Colab/Jupyter notebook.

  `tfds.show_statistics` is a wrapper around
  [tensorflow_data_validation](https://www.tensorflow.org/tfx/data_validation/get_started)
  which calls `tfdv.visualize_statistics`. Statistics are displayed using
  [FACETS OVERVIEW](https://pair-code.github.io/facets/).

  Usage:

  ```
  builder = tfds.builder('mnist')
  tfds.show_statistics(builder.info)
  ```

  Or:

  ```
  ds, ds_info = tfds.load('mnist', with_info)
  tfds.show_statistics(ds_info)
  ```

  Note: In order to work, `tensorflow_data_validation` must be installed and
  the dataset info object must contain the statistics. For "official" datasets,
  only datasets which have been added/updated recently will contains statistics.
  For "custom" datasets, you need to generate the dataset with
  `tensorflow_data_validation` installed to have the statistics.

  Args:
    ds_info: The `tfds.core.DatasetInfo` object containing the statistics.
    split: Split for which generate the statistics.
    disable_logging: `bool`, if True, disable the tfdv logs which can be too
      verbose.

  Returns:
    `None`
  """
  tfdv = lazy_imports_lib.lazy_imports.tensorflow_data_validation

  if split not in ds_info.splits:
    raise ValueError(
        "Invalid requested split: '{}'. Only {} are availables.".format(
            split, list(ds_info.splits)
        )
    )

  # Creates the statistics.
  statistics = statistics_pb2.DatasetFeatureStatisticsList()
  statistics.datasets.add().CopyFrom(ds_info.splits[split].statistics)  # pytype: disable=attribute-error  # bind-properties
  with utils.disable_logging() if disable_logging else utils.nullcontext():
    return tfdv.visualize_statistics(statistics)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common corruptions to images.

Define 15+4 common image corruptions: Gaussian noise, shot noise, impulse_noise,
defocus blur, frosted glass blur, zoom blur, fog, brightness, contrast, elastic,
pixelate, jpeg compression, frost, snow, and motion blur.

4 extra corruptions: gaussian blur, saturate, spatter, and speckle noise.
"""

import io
import subprocess
import tempfile

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds


# To be populated by download_manager
FROST_FILENAMES = []


def _imagemagick_bin():
  return 'imagemagick'  # pylint: disable=unreachable


# /////////////// Corruption Helpers ///////////////


def around_and_astype(x):
  """Round a numpy array, and convert to uint8.

  Args:
    x: numpy array.

  Returns:
    numpy array with dtype uint8.
  """
  return np.around(x).astype(np.uint8)


def disk(radius, alias_blur=0.1, dtype=np.float32):
  """Generating a Gaussian blurring kernel with disk shape.

  Generating a Gaussian blurring kernel with disk shape using cv2 API.

  Args:
    radius: integer, radius of blurring kernel.
    alias_blur: float, standard deviation of Gaussian blurring.
    dtype: data type of kernel

  Returns:
    cv2 object of the Gaussian blurring kernel.
  """
  if radius <= 8:
    length = np.arange(-8, 8 + 1)
    ksize = (3, 3)
  else:
    length = np.arange(-radius, radius + 1)
    ksize = (5, 5)
  x_axis, y_axis = np.meshgrid(length, length)
  aliased_disk = np.array(
      (x_axis**2 + y_axis**2) <= radius**2, dtype=dtype
  )
  aliased_disk /= np.sum(aliased_disk)
  # supersample disk to antialias
  return tfds.core.lazy_imports.cv2.GaussianBlur(
      aliased_disk, ksize=ksize, sigmaX=alias_blur
  )


def clipped_zoom(img, zoom_factor):
  """Zoom image with clipping.

  Zoom the central part of the image and clip extra pixels.

  Args:
    img: numpy array, uncorrupted image.
    zoom_factor: numpy array, a sequence of float numbers for zoom factor.

  Returns:
    numpy array, zoomed image after clipping.
  """
  h = img.shape[0]
  ch = int(np.ceil(h / float(zoom_factor)))
  top_h = (h - ch) // 2

  w = img.shape[1]
  cw = int(np.ceil(w / float(zoom_factor)))
  top_w = (w - cw) // 2

  img = tfds.core.lazy_imports.scipy.ndimage.zoom(
      img[top_h : top_h + ch, top_w : top_w + cw],
      (zoom_factor, zoom_factor, 1),
      order=1,
  )

  # trim off any extra pixels
  trim_top_h = (img.shape[0] - h) // 2
  trim_top_w = (img.shape[1] - w) // 2

  return img[trim_top_h : trim_top_h + h, trim_top_w : trim_top_w + w]


def plasma_fractal(mapsize=512, wibbledecay=3):
  """Generate a heightmap using diamond-square algorithm.

  Modification of the algorithm in
  https://github.com/FLHerne/mapgen/blob/master/diamondsquare.py

  Args:
    mapsize: side length of the heightmap, must be a power of two.
    wibbledecay: integer, decay factor.

  Returns:
    numpy 2d array, side length 'mapsize', of floats in [0,255].
  """
  if mapsize & (mapsize - 1) != 0:
    raise ValueError('mapsize must be a power of two.')
  maparray = np.empty((mapsize, mapsize), dtype=np.float_)
  maparray[0, 0] = 0
  stepsize = mapsize
  wibble = 100

  def wibbledmean(array):
    return array / 4 + wibble * np.random.uniform(-wibble, wibble, array.shape)

  def fillsquares():
    """For each square, calculate middle value as mean of points + wibble."""
    cornerref = maparray[0:mapsize:stepsize, 0:mapsize:stepsize]
    squareaccum = cornerref + np.roll(cornerref, shift=-1, axis=0)
    squareaccum += np.roll(squareaccum, shift=-1, axis=1)
    maparray[
        stepsize // 2 : mapsize : stepsize, stepsize // 2 : mapsize : stepsize
    ] = wibbledmean(squareaccum)

  def filldiamonds():
    """For each diamond, calculate middle value as meanof points + wibble."""
    mapsize = maparray.shape[0]
    drgrid = maparray[
        stepsize // 2 : mapsize : stepsize, stepsize // 2 : mapsize : stepsize
    ]
    ulgrid = maparray[0:mapsize:stepsize, 0:mapsize:stepsize]
    ldrsum = drgrid + np.roll(drgrid, 1, axis=0)
    lulsum = ulgrid + np.roll(ulgrid, -1, axis=1)
    ltsum = ldrsum + lulsum
    maparray[0:mapsize:stepsize, stepsize // 2 : mapsize : stepsize] = (
        wibbledmean(ltsum)
    )
    tdrsum = drgrid + np.roll(drgrid, 1, axis=1)
    tulsum = ulgrid + np.roll(ulgrid, -1, axis=0)
    ttsum = tdrsum + tulsum
    maparray[stepsize // 2 : mapsize : stepsize, 0:mapsize:stepsize] = (
        wibbledmean(ttsum)
    )

  while stepsize >= 2:
    fillsquares()
    filldiamonds()
    stepsize //= 2
    wibble /= wibbledecay

  maparray -= maparray.min()
  return maparray / maparray.max()


# /////////////// End Corruption Helpers ///////////////

# /////////////// Corruptions ///////////////


def gaussian_noise(x, severity=1):
  """Gaussian noise corruption to images.

  Args:
    x: numpy array, uncorrupted image, assumed to have uint8 pixel in [0,255].
    severity: integer, severity of corruption.

  Returns:
    numpy array, image with uint8 pixels in [0,255]. Added Gaussian noise.
  """
  c = [0.08, 0.12, 0.18, 0.26, 0.38][severity - 1]
  x = np.array(x) / 255.0
  x_clip = np.clip(x + np.random.normal(size=x.shape, scale=c), 0, 1) * 255
  return around_and_astype(x_clip)


def shot_noise(x, severity=1):
  """Shot noise corruption to images.

  Args:
    x: numpy array, uncorrupted image, assumed to have uint8 pixel in [0,255].
    severity: integer, severity of corruption.

  Returns:
    numpy array, image with uint8 pixels in [0,255]. Added shot noise.
  """
  c = [60, 25, 12, 5, 3][severity - 1]
  x = np.array(x) / 255.0
  x_clip = np.clip(np.random.poisson(x * c) / float(c), 0, 1) * 255
  return around_and_astype(x_clip)


def impulse_noise(x, severity=1):
  """Impulse noise corruption to images.

  Args:
    x: numpy array, uncorrupted image, assumed to have uint8 pixel in [0,255].
    severity: integer, severity of corruption.

  Returns:
    numpy array, image with uint8 pixels in [0,255]. Added impulse noise.
  """
  c = [0.03, 0.06, 0.09, 0.17, 0.27][severity - 1]
  x = tfds.core.lazy_imports.skimage.util.random_noise(
      np.array(x) / 255.0, mode='s&p', amount=c
  )
  x_clip = np.clip(x, 0, 1) * 255
  return around_and_astype(x_clip)


def defocus_blur(x, severity=1):
  """Defocus blurring to images.

  Apply defocus blurring to images using Gaussian kernel.

  Args:
    x: numpy array, uncorrupted image, assumed to have uint8 pixel in [0,255].
    severity: integer, severity of corruption.

  Returns:
    numpy array, image with uint8 pixels in [0,255]. Applied defocus blur.
  """
  c = [(3, 0.1), (4, 0.5), (6, 0.5), (8, 0.5), (10, 0.5)][severity - 1]
  x = np.array(x) / 255.0
  kernel = disk(radius=c[0], alias_blur=c[1])
  channels = []
  for d in range(3):
    channels.append(tfds.core.lazy_imports.cv2.filter2D(x[:, :, d], -1, kernel))
  channels = np.array(channels).transpose((1, 2, 0))  # 3x224x224 -> 224x224x3
  x_clip = np.clip(channels, 0, 1) * 255
  return around_and_astype(x_clip)


def glass_blur(x, severity=1):
  """Frosted glass blurring to images.

  Apply frosted glass blurring to images by shuffling pixels locally.

  Args:
    x: numpy array, uncorrupted image, assumed to have uint8 pixel in [0,255].
    severity: integer, severity of corruption.

  Returns:
    numpy array, image with uint8 pixels in [0,255]. Applied frosted glass blur.
  """
  # sigma, max_delta, iterations
  c = [(0.7, 1, 2), (0.9, 2, 1), (1, 2, 3), (1.1, 3, 2), (1.5, 4, 2)][
      severity - 1
  ]
  x = np.uint8(
      tfds.core.lazy_imports.skimage.filters.gaussian(
          np.array(x) / 255.0, sigma=c[0], channel_axis=-1
      )
      * 255
  )

  # locally shuffle pixels
  for _ in range(c[2]):
    for h in range(x.shape[0] - c[1], c[1], -1):
      for w in range(x.shape[1] - c[1], c[1], -1):
        dx, dy = np.random.randint(-c[1], c[1], size=(2,))
        h_prime, w_prime = h + dy, w + dx
        # swap
        x[h, w], x[h_prime, w_prime] = x[h_prime, w_prime], x[h, w]
  x_clip = np.clip(
      tfds.core.lazy_imports.skimage.filters.gaussian(
          x / 255.0, sigma=c[0], channel_axis=-1
      ),
      0,
      1,
  )
  x_clip *= 255
  return around_and_astype(x_clip)


def zoom_blur(x, severity=1):
  """Zoom blurring to images.

  Applying zoom blurring to images by zooming the central part of the images.

  Args:
    x: numpy array, uncorrupted image, assumed to have uint8 pixel in [0,255].
    severity: integer, severity of corruption.

  Returns:
    numpy array, image with uint8 pixels in [0,255]. Applied zoom blur.
  """
  c = [
      np.arange(1, 1.11, 0.01),
      np.arange(1, 1.16, 0.01),
      np.arange(1, 1.21, 0.02),
      np.arange(1, 1.26, 0.02),
      np.arange(1, 1.31, 0.03),
  ][severity - 1]
  x = (np.array(x) / 255.0).astype(np.float32)
  out = np.zeros_like(x)
  for zoom_factor in c:
    out += clipped_zoom(x, zoom_factor)
  x = (x + out) / (len(c) + 1)
  x_clip = np.clip(x, 0, 1) * 255
  return around_and_astype(x_clip)


def fog(x, severity=1):
  """Fog corruption to images.

  Adding fog to images. Fog is generated by diamond-square algorithm.

  Args:
    x: numpy array, uncorrupted image, assumed to have uint8 pixel in [0,255].
    severity: integer, severity of corruption.

  Returns:
    numpy array, image with uint8 pixels in [0,255]. Added fog.
  """
  c = [(1.5, 2), (2.0, 2), (2.5, 1.7), (2.5, 1.5), (3.0, 1.4)][severity - 1]
  x = np.array(x) / 255.0
  max_val = x.max()
  mapsize = 512
  shape = x.shape
  max_length = max(shape[0], shape[1])
  if max_length > mapsize:
    mapsize = 2 ** int(np.ceil(np.log2(float(max_length))))
  tmp = plasma_fractal(mapsize=mapsize, wibbledecay=c[1])
  tmp = tmp[: x.shape[0], : x.shape[1]]
  tmp = tmp[..., np.newaxis]
  x += c[0] * tmp
  x_clip = np.clip(x * max_val / (max_val + c[0]), 0, 1) * 255
  return around_and_astype(x_clip)


def brightness(x, severity=1):
  """Change brightness of images.

  Args:
    x: numpy array, uncorrupted image, assumed to have uint8 pixel in [0,255].
    severity: integer, severity of corruption.

  Returns:
    numpy array, image with uint8 pixels in [0,255]. Changed brightness.
  """
  c = [0.1, 0.2, 0.3, 0.4, 0.5][severity - 1]

  x = np.array(x) / 255.0
  x = tfds.core.lazy_imports.skimage.color.rgb2hsv(x)
  x[:, :, 2] = np.clip(x[:, :, 2] + c, 0, 1)
  x = tfds.core.lazy_imports.skimage.color.hsv2rgb(x)
  x_clip = np.clip(x, 0, 1) * 255
  return around_and_astype(x_clip)


def contrast(x, severity=1):
  """Change contrast of images.

  Args:
    x: numpy array, uncorrupted image, assumed to have uint8 pixel in [0,255].
    severity: integer, severity of corruption.

  Returns:
    numpy array, image with uint8 pixels in [0,255]. Changed contrast.
  """
  c = [0.4, 0.3, 0.2, 0.1, 0.05][severity - 1]

  x = np.array(x) / 255.0
  means = np.mean(x, axis=(0, 1), keepdims=True)
  x_clip = np.clip((x - means) * c + means, 0, 1) * 255
  return around_and_astype(x_clip)


def elastic_transform(x, severity=1):
  """Conduct elastic transform to images.

  Elastic transform is performed on small patches of the images.

  Args:
    x: numpy array, uncorrupted image, assumed to have uint8 pixel in [0,255].
    severity: integer, severity of corruption.

  Returns:
    numpy array, image with uint8 pixels in [0,255]. Applied elastic transform.
  """
  c = [
      (244 * 2, 244 * 0.7, 244 * 0.1),
      (244 * 2, 244 * 0.08, 244 * 0.2),
      (244 * 0.05, 244 * 0.01, 244 * 0.02),
      (244 * 0.07, 244 * 0.01, 244 * 0.02),
      (244 * 0.12, 244 * 0.01, 244 * 0.02),
  ][severity - 1]

  image = np.array(x, dtype=np.float32) / 255.0
  shape = image.shape
  shape_size = shape[:2]

  # random affine
  center_square = np.float32(shape_size) // 2
  square_size = min(shape_size) // 3
  pts1 = np.float32([
      center_square + square_size,
      [center_square[0] + square_size, center_square[1] - square_size],
      center_square - square_size,
  ])
  pts2 = pts1 + np.random.uniform(-c[2], c[2], size=pts1.shape).astype(
      np.float32
  )
  affine_trans = tfds.core.lazy_imports.cv2.getAffineTransform(pts1, pts2)
  image = tfds.core.lazy_imports.cv2.warpAffine(
      image,
      affine_trans,
      shape_size[::-1],
      borderMode=tfds.core.lazy_imports.cv2.BORDER_REFLECT_101,
  )

  dx = (
      tfds.core.lazy_imports.skimage.filters.gaussian(
          np.random.uniform(-1, 1, size=shape[:2]),
          c[1],
          mode='reflect',
          truncate=3,
      )
      * c[0]
  ).astype(np.float32)
  dy = (
      tfds.core.lazy_imports.skimage.filters.gaussian(
          np.random.uniform(-1, 1, size=shape[:2]),
          c[1],
          mode='reflect',
          truncate=3,
      )
      * c[0]
  ).astype(np.float32)
  dx, dy = dx[..., np.newaxis], dy[..., np.newaxis]

  x, y, z = np.meshgrid(
      np.arange(shape[1]), np.arange(shape[0]), np.arange(shape[2])
  )
  indices = (
      np.reshape(y + dy, (-1, 1)),
      np.reshape(x + dx, (-1, 1)),
      np.reshape(z, (-1, 1)),
  )
  x_clip = (
      np.clip(
          tfds.core.lazy_imports.scipy.ndimage.interpolation.map_coordinates(
              image, indices, order=1, mode='reflect'
          ).reshape(shape),
          0,
          1,
      )
      * 255
  )
  return around_and_astype(x_clip)


def pixelate(x, severity=1):
  """Pixelate images.

  Conduct pixelating corruptions to images by first shrinking the images and
  then resizing to original size.

  Args:
    x: numpy array, uncorrupted image, assumed to have uint8 pixel in [0,255].
    severity: integer, severity of corruption.

  Returns:
    numpy array, image with uint8 pixels in [0,255]. Applied pixelating
    corruption.
  """
  c = [0.6, 0.5, 0.4, 0.3, 0.25][severity - 1]
  shape = x.shape
  x = tfds.core.lazy_imports.PIL_Image.fromarray(x.astype(np.uint8))
  x = x.resize((int(shape[1] * c), int(shape[0] * c)))
  x = x.resize((shape[1], shape[0]))
  return np.asarray(x)


def jpeg_compression(x, severity=1):
  """Conduct jpeg compression to images.

  Args:
    x: numpy array, uncorrupted image, assumed to have uint8 pixel in [0,255].
    severity: integer, severity of corruption.

  Returns:
    numpy array, image with uint8 pixels in [0,255]. Applied jpeg compression.
  """
  c = [25, 18, 15, 10, 7][severity - 1]
  x = tfds.core.lazy_imports.PIL_Image.fromarray(x.astype(np.uint8))
  output = io.BytesIO()
  x.save(output, 'JPEG', quality=c)
  output.seek(0)
  x = tfds.core.lazy_imports.PIL_Image.open(output)
  return np.asarray(x)


def frost(x, severity=1):
  """Apply frost to images.

  Args:
    x: numpy array, uncorrupted image, assumed to have uint8 pixel in [0,255].
    severity: integer, severity of corruption.

  Returns:
    numpy array, image with uint8 pixels in [0,255]. Applied frost.
  """
  c = [(1, 0.4), (0.8, 0.6), (0.7, 0.7), (0.65, 0.7), (0.6, 0.75)][severity - 1]
  filename = FROST_FILENAMES[np.random.randint(5)]
  with tempfile.NamedTemporaryFile() as im_frost:
    tf.io.gfile.copy(filename, im_frost.name, overwrite=True)
    frost_img = tfds.core.lazy_imports.cv2.imread(im_frost.name)
  # randomly crop and convert to rgb
  x_start, y_start = np.random.randint(
      0, frost_img.shape[0] - 224
  ), np.random.randint(0, frost_img.shape[1] - 224)
  frost_img = frost_img[x_start : x_start + 224, y_start : y_start + 224][
      ..., [2, 1, 0]
  ]

  x = np.clip(c[0] * np.array(x) + c[1] * frost_img, 0, 255)

  return around_and_astype(x)


def snow(x, severity=1):
  """Apply snow to images.

  Args:
    x: numpy array, uncorrupted image, assumed to have uint8 pixel in [0,255].
    severity: integer, severity of corruption.

  Returns:
    numpy array, image with uint8 pixels in [0,255]. Applied snow.
  """
  cv2 = tfds.core.lazy_imports.cv2
  PIL_Image = tfds.core.lazy_imports.PIL_Image  # pylint: disable=invalid-name
  c = [
      (0.1, 0.3, 3, 0.5, 10, 4, 0.8),
      (0.2, 0.3, 2, 0.5, 12, 4, 0.7),
      (0.55, 0.3, 4, 0.9, 12, 8, 0.7),
      (0.55, 0.3, 4.5, 0.85, 12, 8, 0.65),
      (0.55, 0.3, 2.5, 0.85, 12, 12, 0.55),
  ][severity - 1]

  x = np.array(x, dtype=np.float32) / 255.0
  snow_layer = np.random.normal(
      size=x.shape[:2], loc=c[0], scale=c[1]
  )  # [:2] for monochrome

  snow_layer = clipped_zoom(snow_layer[..., np.newaxis], c[2])
  snow_layer[snow_layer < c[3]] = 0

  snow_layer = PIL_Image.fromarray(
      (np.clip(snow_layer.squeeze(), 0, 1) * 255).astype(np.uint8), mode='L'
  )

  with tempfile.NamedTemporaryFile() as im_input:
    with tempfile.NamedTemporaryFile() as im_output:
      snow_layer.save(im_input.name, format='PNG')

      convert_bin = _imagemagick_bin()
      radius = c[4]
      sigma = c[5]
      angle = np.random.uniform(-135, -45)

      subprocess.check_output([
          convert_bin,
          '-motion-blur',
          '{}x{}+{}'.format(radius, sigma, angle),
          im_input.name,
          im_output.name,
      ])

      with open(im_output.name, 'rb') as f:
        output = f.read()

  snow_layer = (
      cv2.imdecode(np.frombuffer(output, np.uint8), cv2.IMREAD_UNCHANGED)
      / 255.0
  )
  snow_layer = snow_layer[..., np.newaxis]

  x = c[6] * x + (1 - c[6]) * np.maximum(
      x, cv2.cvtColor(x, cv2.COLOR_RGB2GRAY).reshape(224, 224, 1) * 1.5 + 0.5
  )
  x = np.clip(x + snow_layer + np.rot90(snow_layer, k=2), 0, 1) * 255

  return around_and_astype(x)


def motion_blur(x, severity=1):
  """Apply motion blur to images.

  Args:
    x: numpy array, uncorrupted image, assumed to have uint8 pixel in [0,255].
    severity: integer, severity of corruption.

  Returns:
    numpy array, image with uint8 pixels in [0,255]. Applied motion blur.
  """
  c = [(10, 3), (15, 5), (15, 8), (15, 12), (20, 15)][severity - 1]

  x = tfds.core.lazy_imports.PIL_Image.fromarray(x.astype(np.uint8))

  with tempfile.NamedTemporaryFile() as im_input:
    with tempfile.NamedTemporaryFile() as im_output:
      x.save(im_input.name, format='PNG')

      convert_bin = _imagemagick_bin()
      radius = c[0]
      sigma = c[1]
      angle = np.random.uniform(-45, -45)

      subprocess.check_output([
          convert_bin,
          '-motion-blur',
          '{}x{}+{}'.format(radius, sigma, angle),
          im_input.name,
          im_output.name,
      ])

      with open(im_output.name, 'rb') as f:
        output = f.read()

  x = tfds.core.lazy_imports.cv2.imdecode(
      np.frombuffer(output, np.uint8),
      tfds.core.lazy_imports.cv2.IMREAD_UNCHANGED,
  )

  if x.shape != (224, 224):
    x = np.clip(x[..., [2, 1, 0]], 0, 255)  # BGR to RGB
  else:  # greyscale to RGB
    x = np.clip(np.array([x, x, x]).transpose((1, 2, 0)), 0, 255)

  return around_and_astype(x)


# /////////////// Extra Corruptions ///////////////


def gaussian_blur(x, severity=1):
  """Apply gaussian blur to images.

  Args:
    x: numpy array, uncorrupted image, assumed to have uint8 pixel in [0,255].
    severity: integer, severity of corruption.

  Returns:
    numpy array, image with uint8 pixels in [0,255]. Applied gaussian blur.
  """
  c = [1, 2, 3, 4, 6][severity - 1]

  x = tfds.core.lazy_imports.skimage.filters.gaussian(
      np.array(x) / 255.0, sigma=c, channel_axis=-1
  )
  x = np.clip(x, 0, 1) * 255

  return around_and_astype(x)


def saturate(x, severity=1):
  """Increase saturation of images.

  Args:
    x: numpy array, uncorrupted image, assumed to have uint8 pixel in [0,255].
    severity: integer, severity of corruption.

  Returns:
    numpy array, image with uint8 pixels in [0,255]. Applied saturation.
  """
  c = [(0.3, 0), (0.1, 0), (2, 0), (5, 0.1), (20, 0.2)][severity - 1]

  x = np.array(x) / 255.0
  x = tfds.core.lazy_imports.skimage.color.rgb2hsv(x)
  x[:, :, 1] = np.clip(x[:, :, 1] * c[0] + c[1], 0, 1)
  x = tfds.core.lazy_imports.skimage.color.hsv2rgb(x)
  x = np.clip(x, 0, 1) * 255

  return around_and_astype(x)


def spatter(x, severity=1):
  """Apply spatter to images.

  Args:
    x: numpy array, uncorrupted image, assumed to have uint8 pixel in [0,255].
    severity: integer, severity of corruption.

  Returns:
    numpy array, image with uint8 pixels in [0,255]. Applied spatter.
  """
  cv2 = tfds.core.lazy_imports.cv2
  skimage = tfds.core.lazy_imports.skimage
  c = [
      (0.65, 0.3, 4, 0.69, 0.6, 0),
      (0.65, 0.3, 3, 0.68, 0.6, 0),
      (0.65, 0.3, 2, 0.68, 0.5, 0),
      (0.65, 0.3, 1, 0.65, 1.5, 1),
      (0.67, 0.4, 1, 0.65, 1.5, 1),
  ][severity - 1]
  x = np.array(x, dtype=np.float32) / 255.0

  liquid_layer = np.random.normal(size=x.shape[:2], loc=c[0], scale=c[1])

  liquid_layer = skimage.filters.gaussian(liquid_layer, sigma=c[2])
  liquid_layer[liquid_layer < c[3]] = 0
  if c[5] == 0:
    liquid_layer = (liquid_layer * 255).astype(np.uint8)
    dist = 255 - cv2.Canny(liquid_layer, 50, 150)
    dist = cv2.distanceTransform(dist, cv2.DIST_L2, 5)
    _, dist = cv2.threshold(dist, 20, 20, cv2.THRESH_TRUNC)
    dist = cv2.blur(dist, (3, 3)).astype(np.uint8)
    dist = cv2.equalizeHist(dist)
    #     ker = np.array([[-1,-2,-3],[-2,0,0],[-3,0,1]], dtype=np.float32)
    #     ker -= np.mean(ker)
    ker = np.array([[-2, -1, 0], [-1, 1, 1], [0, 1, 2]])
    dist = cv2.filter2D(dist, cv2.CV_8U, ker)
    dist = cv2.blur(dist, (3, 3)).astype(np.float32)

    m = cv2.cvtColor(liquid_layer * dist, cv2.COLOR_GRAY2BGRA)
    m /= np.max(m, axis=(0, 1))
    m *= c[4]

    # water is pale turqouise
    color = np.concatenate(
        (
            175 / 255.0 * np.ones_like(m[..., :1]),
            238 / 255.0 * np.ones_like(m[..., :1]),
            238 / 255.0 * np.ones_like(m[..., :1]),
        ),
        axis=2,
    )

    color = cv2.cvtColor(color, cv2.COLOR_BGR2BGRA)
    x = cv2.cvtColor(x, cv2.COLOR_BGR2BGRA)

    x = cv2.cvtColor(np.clip(x + m * color, 0, 1), cv2.COLOR_BGRA2BGR) * 255
  else:
    m = np.where(liquid_layer > c[3], 1, 0)
    m = skimage.filters.gaussian(m.astype(np.float32), sigma=c[4])
    m[m < 0.8] = 0
    #         m = np.abs(m) ** (1/c[4])

    # mud brown
    color = np.concatenate(
        (
            63 / 255.0 * np.ones_like(x[..., :1]),
            42 / 255.0 * np.ones_like(x[..., :1]),
            20 / 255.0 * np.ones_like(x[..., :1]),
        ),
        axis=2,
    )

    color *= m[..., np.newaxis]
    x *= 1 - m[..., np.newaxis]

    x = np.clip(x + color, 0, 1) * 255
  return around_and_astype(x)


def speckle_noise(x, severity=1):
  """Apply speckle noise to images.

  Args:
    x: numpy array, uncorrupted image, assumed to have uint8 pixel in [0,255].
    severity: integer, severity of corruption.

  Returns:
    numpy array, image with uint8 pixels in [0,255]. Applied speckle noise.
  """
  c = [0.15, 0.2, 0.35, 0.45, 0.6][severity - 1]

  x = np.array(x) / 255.0
  x = np.clip(x + x * np.random.normal(size=x.shape, scale=c), 0, 1) * 255
  return around_and_astype(x)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""To write records into sharded records files."""

import dataclasses
import functools
import itertools
import json
import os
from typing import Any, Iterable, List, Optional, Tuple

from absl import logging
from etils import epath
from tensorflow_datasets.core import example_parser
from tensorflow_datasets.core import example_serializer
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core import hashing
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import shuffle
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.utils import file_utils
from tensorflow_datasets.core.utils import shard_utils
from tensorflow_datasets.core.utils import type_utils


# TODO(tfds): Should be `TreeDict[FeatureValue]`
Example = Any

_INDEX_PATH_SUFFIX = "_index.json"


@dataclasses.dataclass(frozen=True)
class _ShardSpec:
  """Spec to write a final records shard.

  Attributes:
    shard_index: Index of the shard.
    path: The path where to write the shard.
    index_path: The path where to write index of the records in the
      corresponding shard. NOTE: Value for this attribute is always set, but
      usage depends on whether `write_examples` returned a list of record
      positions for each example.
    examples_number: Number of examples in shard.
    file_instructions: Reading instructions.
  """

  shard_index: int
  path: str
  index_path: str
  examples_number: int
  file_instructions: List[shard_utils.FileInstruction]


def _raise_error_for_duplicated_keys(example1, example2, example_specs):
  """Log information about the examples and raise an AssertionError."""
  msg = "Two examples share the same hashed key!"
  logging.error(msg)
  parser = example_parser.ExampleParser(example_specs)
  ex1 = parser.parse_example(example1)
  ex2 = parser.parse_example(example2)
  logging.error("1st example: %s", ex1)
  logging.error("2nd example: %s", ex2)
  raise AssertionError(msg + " See logs above to view the examples.")


def _get_index_path(path: str) -> epath.PathLike:
  """Returns path to the index file of the records stored at the given path.

  E.g: Say the path to a shard of records (of a particular split) are stored at
  `your/path/to/records/foo.riegeli-00001-of-00005`, it is transformed into
  `your/path/to/records/foo.riegeli-00001-of-00005_index.json`.

  Args:
    path: Path to the record file.

  Returns:
    Path of the index file of a shard of records stored at given path.
  """
  return path + _INDEX_PATH_SUFFIX


def _get_shard_specs(
    num_examples: int,
    total_size: int,
    bucket_lengths: List[int],
    filename_template: naming.ShardedFileTemplate,
    shard_config: shard_utils.ShardConfig,
) -> List[_ShardSpec]:
  """Returns list of _ShardSpec instances, corresponding to shards to write.

  Args:
    num_examples: int, number of examples in split.
    total_size: int (bytes), sum of example sizes.
    bucket_lengths: list of ints, number of examples in each bucket.
    filename_template: template to format sharded filenames.
    shard_config: the configuration for creating shards.
  """
  num_shards = shard_config.get_number_shards(total_size, num_examples)
  shard_boundaries = _get_shard_boundaries(num_examples, num_shards)
  shard_specs = []
  bucket_indexes = [str(i) for i in range(len(bucket_lengths))]
  from_ = 0
  for shard_index, to in enumerate(shard_boundaries):
    # Read the bucket indexes
    file_instructions = shard_utils.get_file_instructions(
        from_, to, bucket_indexes, bucket_lengths
    )
    shard_path = filename_template.sharded_filepath(
        shard_index=shard_index, num_shards=num_shards
    )
    index_path = _get_index_path(os.fspath(shard_path))
    shard_specs.append(
        _ShardSpec(
            shard_index=shard_index,
            path=os.fspath(shard_path),
            index_path=index_path,
            examples_number=to - from_,
            file_instructions=file_instructions,
        )
    )
    from_ = to
  return shard_specs


def _get_shard_boundaries(
    num_examples: int,
    number_of_shards: int,
) -> List[int]:
  if num_examples == 0:
    raise AssertionError("No examples were yielded.")
  if num_examples < number_of_shards:
    raise AssertionError(
        "num_examples ({}) < number_of_shards ({})".format(
            num_examples, number_of_shards
        )
    )
  return [
      int(round(num_examples * (float(i) / number_of_shards)))
      for i in range(1, number_of_shards + 1)
  ]


def _write_examples(
    path: epath.PathLike,
    iterator: Iterable[type_utils.KeySerializedExample],
    file_format: file_adapters.FileFormat = file_adapters.DEFAULT_FILE_FORMAT,
) -> Optional[file_adapters.ExamplePositions]:
  """Write examples from iterator in the given `file_format`."""
  return file_adapters.ADAPTER_FOR_FORMAT[file_format].write_examples(
      path, iterator
  )


def _write_index_file(
    sharded_index_path: epath.PathLike, record_keys: List[Any]
):
  """Writes index file for records of a shard at given `sharded_index_path`.

  NOTE: Each record position (i.e shard_key in shard_keys) is stored as a
  string for better readability of the index files. The reader should parse the
  position from string using `RecordPosition.from_str()` method.

  Args:
    sharded_index_path: Path to the sharded index path.
    record_keys: List of keys/indices of the records in the shard.
  """
  # Store string representation of each record position.
  #
  # NOTE: This makes the index file more readable. Although the reader should
  # parse the record position from a string.
  index_info = {"index": [str(record_key) for record_key in record_keys]}
  epath.Path(sharded_index_path).write_text(json.dumps(index_info))
  logging.info("Wrote index file to %s", os.fspath(sharded_index_path))


class Writer(object):
  """Shuffles and writes Examples to sharded TFRecord files.

  The number of shards is computed automatically.
  """

  def __init__(
      self,
      serializer: example_serializer.Serializer,
      filename_template: naming.ShardedFileTemplate,
      hash_salt,
      disable_shuffling: bool,
      file_format=file_adapters.DEFAULT_FILE_FORMAT,
      shard_config: Optional[shard_utils.ShardConfig] = None,
  ):
    """Initializes Writer.

    Args:
      serializer: class that can serialize examples.
      filename_template: template to format sharded filenames.
      hash_salt (str or bytes): salt to hash keys.
      disable_shuffling (bool): Specifies whether to shuffle the records.
      file_format (FileFormat): format of the record files in which the dataset
        should be written in.
      shard_config: the configuration for creating shards.
    """
    self._serializer = serializer
    self._shuffler = shuffle.Shuffler(
        dirpath=filename_template.data_dir,
        hash_salt=hash_salt,
        disable_shuffling=disable_shuffling,
    )
    self._num_examples = 0
    self._filename_template = filename_template
    self._file_format = file_format
    self._shard_config = shard_config or shard_utils.ShardConfig()

  def write(self, key, example):
    """Writes given Example.

    The given example is not directly written to the tfrecord file, but to a
    temporary file (or memory). The finalize() method does write the tfrecord
    files.

    Args:
      key (int|bytes): the key associated with the example. Used for shuffling.
      example: the Example to write to the tfrecord file.
    """
    serialized_example = self._serializer.serialize_example(example=example)
    self._shuffler.add(key, serialized_example)
    self._num_examples += 1

  def finalize(self) -> Tuple[List[int], int]:
    """Effectively writes examples to the tfrecord files."""
    filename = self._filename_template.sharded_filepaths_pattern()
    shard_specs = _get_shard_specs(
        self._num_examples,
        self._shuffler.size,
        self._shuffler.bucket_lengths,
        self._filename_template,
        self._shard_config,
    )
    # Here we just loop over the examples, and don't use the instructions, just
    # the final number of examples in every shard. Instructions could be used to
    # parallelize, but one would need to be careful not to sort buckets twice.
    examples_generator = iter(
        utils.tqdm(
            self._shuffler,
            desc=f"Shuffling {filename}...",
            total=self._num_examples,
            unit=" examples",
            leave=False,
        )
    )
    try:
      for shard_spec in shard_specs:
        iterator = itertools.islice(
            examples_generator, 0, shard_spec.examples_number
        )
        record_keys = _write_examples(
            shard_spec.path, iterator, self._file_format
        )

        # No shard keys returned (e.g: TFRecord format), index cannot be
        # created.
        if not record_keys:
          continue

        # Number of `shard_keys` received should match the number of examples
        # written in this shard.
        if len(record_keys) != int(shard_spec.examples_number):
          raise RuntimeError(
              f"Length of example `keys` ({len(record_keys)}) does not match "
              f"`shard_spec.examples_number: (`{shard_spec.examples_number})"
          )
        _write_index_file(shard_spec.index_path, record_keys)

    except shuffle.DuplicatedKeysError as err:
      _raise_error_for_duplicated_keys(
          err.item1, err.item2, self._serializer.example_specs
      )

    # Finalize the iterator to clear-up TQDM
    try:
      val = next(examples_generator)
    except StopIteration:
      pass
    else:
      raise ValueError(
          f"Shuffling more elements than expected. Additional element: {val}"
      )

    shard_lengths = [int(spec.examples_number) for spec in shard_specs]

    logging.info(
        "Done writing %s. Number of examples: %s (shards: %s)",
        filename,
        sum(shard_lengths),
        shard_lengths,
    )
    return shard_lengths, self._shuffler.size


@dataclasses.dataclass
class _ShardInfo:
  id: int
  num_examples: int
  size: int


class BeamWriter(object):
  """Shuffles / writes Examples beam collection to sharded TFRecord files.

  The given examples are not directly writen to the final tfrecord file, but to
  temporary files.
  """

  _OUTPUT_TAG_BUCKETS_LEN_SIZE = "tag_buckets_len_size"

  def __init__(
      self,
      serializer: example_serializer.Serializer,
      filename_template: naming.ShardedFileTemplate,
      hash_salt,
      disable_shuffling: bool,
      file_format: file_adapters.FileFormat = file_adapters.DEFAULT_FILE_FORMAT,
      shard_config: Optional[shard_utils.ShardConfig] = None,
  ):
    """Init BeamWriter.

    Note that file "{filepath_prefix}.shard_lengths.json" is also created. It
    contains a list with the number of examples in each final shard. Eg:
    "[10,11,10,11]".

    Args:
      serializer: class that can serialize examples.
      filename_template: template to format sharded filenames.
      hash_salt: string, the salt to use for hashing of keys.
      disable_shuffling: bool, specifies whether to shuffle the records.
      file_format: file_adapters.FileFormat, format of the record files in which
        the dataset will be read/written from.
      shard_config: the configuration for creating shards.
    """
    self._original_state = dict(
        serializer=serializer,
        filename_template=filename_template,
        hash_salt=hash_salt,
        disable_shuffling=disable_shuffling,
        file_format=file_format,
        shard_config=shard_config,
    )
    self._filename_template = filename_template
    self._split_info_path = (
        f"{filename_template.filepath_prefix()}.split_info.json"
    )
    self._serializer = serializer
    self._hasher = hashing.Hasher(hash_salt)
    self._split_info = None
    self._file_format = file_format
    self._disable_shuffling = disable_shuffling
    self._shard_config = shard_config or shard_utils.ShardConfig()

  @functools.lru_cache()
  def _get_counter(self, name: str, namespace: str = "BeamWriter"):
    beam = lazy_imports_lib.lazy_imports.apache_beam
    return beam.metrics.Metrics.counter(namespace, name)

  def inc_counter(self, name: str, value: int = 1) -> None:
    self._get_counter(name).inc(value)

  def __getstate__(self):
    return self._original_state

  def __setstate__(self, state):
    self.__init__(**state)

  def _serialize_example(
      self,
      key_example: Tuple[hashing.HashKey, Example],
  ) -> Tuple[Any, bytes]:
    """Returns (shard#, (hkey, serialized_example))."""
    key, example = key_example
    serialized_example = self._serializer.serialize_example(example)
    if self._disable_shuffling:
      hkey = key
    else:
      hkey = self._hasher.hash_key(key)
    self.inc_counter(name="serialized_examples")
    return (hkey, serialized_example)

  def _check_num_examples(self, num_examples: int) -> int:
    if num_examples <= 0:
      raise ValueError(
          f"The total number of generated examples is {num_examples}. This"
          " should be >0!"
      )
    return num_examples

  def _write_final_shard(
      self,
      shardid_examples: Tuple[int, Iterable[type_utils.KeySerializedExample]],
      non_empty_shard_ids: List[int],
  ) -> _ShardInfo:
    """Write all examples of a shard to disk.

    Arguments:
      shardid_examples: tuple of the shard id and the serialized examples that
        belong to that shard.
      non_empty_shard_ids: list of the shard ids of all the non-empty shards.
        Must be sorted.

    Returns:
      the shard info of the written shard.
    """
    original_shard_id, examples = shardid_examples
    if not examples:
      raise AssertionError("Not a single example present in the PCollection!")
    # There may be empty shards, this ensures there are no gaps.
    shard_id = non_empty_shard_ids.index(original_shard_id)
    examples = sorted(examples)
    # Compare continuous examples
    for ex0, ex1 in zip(examples[:-1], examples[1:]):
      if ex0[0] == ex1[0]:  # Different keys
        _raise_error_for_duplicated_keys(
            ex0[1], ex1[1], self._serializer.example_specs
        )
    shard_path = self._filename_template.sharded_filepath(
        shard_index=shard_id, num_shards=len(non_empty_shard_ids)
    )
    with utils.incomplete_file(epath.Path(shard_path)) as tmp_path:
      logging.info(
          "Writing %d examples to %s.", len(examples), os.fspath(tmp_path)
      )
      record_keys = _write_examples(tmp_path, examples, self._file_format)
    self.inc_counter(name="written_shards")
    # If there are record_keys, create index files.
    if record_keys:
      index_path = _get_index_path(os.fspath(shard_path))
      _write_index_file(index_path, list(record_keys))
    shard_size = sum(map(len, examples))
    return _ShardInfo(id=shard_id, num_examples=len(examples), size=shard_size)

  def _number_of_shards(self, num_examples: int, total_size: int) -> int:
    """Returns the number of shards."""
    num_shards = self._shard_config.get_number_shards(
        total_size=total_size,
        num_examples=num_examples,
        uses_precise_sharding=False,
    )
    self.inc_counter(name="NumberOfShards", value=num_shards)
    return num_shards

  def _assign_shard(
      self,
      key_serialized_example: Tuple[Any, bytes],
      num_shards: int,
      largest_key: List[int],
  ) -> Tuple[int, Tuple[Any, bytes]]:
    """Assigns a shard id to the example."""
    key, _ = key_serialized_example
    largest_key = largest_key[0]
    shard_number = shuffle.get_bucket_number(
        hkey=key, num_buckets=num_shards, max_hkey=largest_key
    )
    return (shard_number, key_serialized_example)

  def _store_split_info(
      self,
      shard_infos: List[_ShardInfo],
      total_size: int,
  ) -> None:
    """Stores the split info to disk."""
    shard_infos = sorted(shard_infos, key=lambda x: x.id)
    shard_lengths = [info.num_examples for info in shard_infos]
    with utils.incomplete_file(epath.Path(self._split_info_path)) as tmp_path:
      tmp_path.write_text(
          json.dumps({"total_size": total_size, "shard_lengths": shard_lengths})
      )

  def _sort_shard_ids(
      self, shard_ids: List[int], ideal_num_shards: int
  ) -> List[int]:
    """Returns the sorted shard ids and logs information."""
    if len(shard_ids) != ideal_num_shards:
      logging.info(
          "Ideally there would be %d shards, but got %d non-empty shards.",
          ideal_num_shards,
          len(shard_ids),
      )
    return sorted(shard_ids)

  def write_from_pcollection(self, examples_pcollection):
    """Returns PTransform to write (key, example) PCollection to tfrecords."""
    beam = lazy_imports_lib.lazy_imports.apache_beam
    serialized_examples = (
        examples_pcollection
        | "Serialize" >> beam.Map(self._serialize_example)
        # (key, serialized_example)
    )

    largest_key = (
        serialized_examples
        | beam.Keys()
        | "LargestKey" >> beam.combiners.Top.Largest(1)
    )
    num_examples = (
        serialized_examples
        | "CountExamples" >> beam.combiners.Count.Globally()
        | "CheckValidNumExamples" >> beam.Map(self._check_num_examples)
    )
    total_size = beam.pvalue.AsSingleton(
        serialized_examples
        | beam.Values()
        | beam.Map(len)
        | "TotalSize" >> beam.CombineGlobally(sum)
    )
    ideal_num_shards = beam.pvalue.AsSingleton(
        num_examples
        | "NumberOfShards"
        >> beam.Map(self._number_of_shards, total_size=total_size)
    )

    examples_per_shard = (
        serialized_examples
        | "AssignShard"
        >> beam.Map(
            self._assign_shard,
            largest_key=beam.pvalue.AsSingleton(largest_key),
            num_shards=ideal_num_shards,
        )
        # (shard_id, serialized_example)
        | "GroupShards" >> beam.GroupByKey()
        # (shard_id, [serialized_example])
    )

    # There may be shards that did not get assigned any examples. We will ignore
    # those shards and for that we need the ids of shards that are non-empty.
    non_empty_shard_ids = beam.pvalue.AsSingleton(
        examples_per_shard
        | "GetIdsOfNonEmptyShards" >> beam.Keys()
        | "CollectIdsOfNonEmptyShards" >> beam.transforms.combiners.ToList()
        | "SortIdsOfNonEmptyShards"
        >> beam.Map(self._sort_shard_ids, ideal_num_shards=ideal_num_shards)
    )

    return (
        examples_per_shard
        # (shard_id, [serialized_example])
        | "WriteFinalShards"
        >> beam.Map(
            self._write_final_shard, non_empty_shard_ids=non_empty_shard_ids
        )
        # (_ShardInfo)
        | "CollectShardInfo" >> beam.transforms.combiners.ToList()
        # [_ShardInfo]
        | "CalculateSplitInfo"
        >> beam.ParDo(self._store_split_info, total_size=total_size)
    )

  def finalize(self) -> Tuple[List[int], int]:
    """Deletes tmp directory and returns shard_lengths and total_size.

    Returns:
      List of length <number of shards> containing the number of examples stored
      in each shard, and size of the files (in bytes).
    """
    if self._split_info is None:
      split_info_path = epath.Path(self._split_info_path)
      self._split_info = json.loads(split_info_path.read_bytes())
      split_info_path.unlink()
    return self._split_info["shard_lengths"], self._split_info["total_size"]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Video feature."""

from __future__ import annotations

import os
import tempfile
from typing import Optional, Sequence, Union

from etils import epath
import numpy as np
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.features import feature as feature_lib
from tensorflow_datasets.core.features import image_feature
from tensorflow_datasets.core.features import sequence_feature
from tensorflow_datasets.core.proto import feature_pb2
from tensorflow_datasets.core.utils import type_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

Json = type_utils.Json


class Video(sequence_feature.Sequence):
  """`FeatureConnector` for videos, encoding frames individually on disk.

  Video: The image connector accepts as input a 4 dimensional `tf.uint8` array
  representing a video, a sequence of paths to encoded frames, or a path or a
  file object that can be decoded with ffmpeg. Note that not all formats in
  ffmpeg support reading from pipes, so providing a file object might fail.
  Furthermore, if a path is given that is not on the local file system, we first
  copy it to a temporary local file before passing it to ffmpeg.

  Output:
    video: tf.Tensor of type `tf.uint8` and shape
      [num_frames, height, width, channels], where channels must be 1 or 3

  Example:

    * In the DatasetInfo object:

    ```
    features=features.FeatureDict({
        'video': features.Video(shape=(None, 64, 64, 3)),
    })
    ```

    * During generation, you can use any of:

    ```
    yield {
        'video': np.ones(shape=(128, 64, 64, 3), dtype=np.uint8),
    }
    ```

    or list of frames:

    ```
    yield {
        'video': ['path/to/frame001.png', 'path/to/frame002.png'],
    }
    ```

    or path to video (including `os.PathLike` objects):

    ```
    yield {
        'video': '/path/to/video.avi',
    }
    ```

    or file object (or `bytes`):

    ```
    yield {
        'video': tf.io.gfile.GFile('/complex/path/video.avi'),
    }
    ```
  """

  def __init__(
      self,
      shape: Sequence[Optional[int]],
      encoding_format: str = 'png',
      ffmpeg_extra_args: Sequence[str] = (),
      use_colormap: bool = False,
      dtype: Union[np.uint8, tf.uint8, np.uint16, tf.uint16] = np.uint8,
      doc: feature_lib.DocArg = None,
  ):
    """Initializes the connector.

    Args:
      shape: tuple of ints, the shape of the video (num_frames, height, width,
        channels), where channels is 1 or 3.
      encoding_format: The video is stored as a sequence of encoded images. You
        can use any encoding format supported by image_feature.Feature.
      ffmpeg_extra_args: A sequence of additional args to be passed to the
        ffmpeg binary. Specifically, ffmpeg will be called as: `` ffmpeg -i
        <input_file> <ffmpeg_extra_args> %010d.<encoding_format> ``
      use_colormap: Forwarded to `tfds.features.Image`. If `True`,
        `tfds.as_dataframe` will display each value in the image with a
        different color.
      dtype: tf.uint16 or tf.uint8 (default). tf.uint16 can be used only with
        png encoding_format
      doc: Documentation of this feature (e.g. description).

    Raises:
      ValueError: If the shape is invalid
    """
    dtype = tf.dtypes.as_dtype(dtype)
    shape = tuple(shape)
    if len(shape) != 4:
      raise ValueError('Video shape should be of rank 4')
    self._encoding_format = encoding_format
    self._extra_ffmpeg_args = list(ffmpeg_extra_args or [])
    super(Video, self).__init__(
        image_feature.Image(
            shape=shape[1:],
            dtype=dtype,
            encoding_format=encoding_format,
            use_colormap=use_colormap,
        ),
        length=shape[0],
    )

  def _ffmpeg_decode(self, path_or_fobj):
    if isinstance(path_or_fobj, epath.PathLikeCls):
      ffmpeg_args = ['-i', os.fspath(path_or_fobj)]
      ffmpeg_stdin = None
    else:
      ffmpeg_args = ['-i', 'pipe:0']
      ffmpeg_stdin = path_or_fobj.read()
    ffmpeg_args += self._extra_ffmpeg_args

    with tempfile.TemporaryDirectory() as ffmpeg_dir:
      out_pattern = os.path.join(ffmpeg_dir, f'%010d.{self._encoding_format}')
      ffmpeg_args.append(out_pattern)
      utils.ffmpeg_run(ffmpeg_args, ffmpeg_stdin)
      frames = [  # Load all encoded images
          p.read_bytes() for p in sorted(epath.Path(ffmpeg_dir).iterdir())
      ]
    return frames

  def encode_example(self, video_or_path_or_fobj):
    """Converts the given image into a dict convertible to tf example."""
    if isinstance(video_or_path_or_fobj, epath.PathLikeCls):
      video_or_path_or_fobj = os.fspath(video_or_path_or_fobj)
      if not os.path.isfile(video_or_path_or_fobj):
        _, video_temp_path = tempfile.mkstemp()
        try:
          tf.io.gfile.copy(
              video_or_path_or_fobj, video_temp_path, overwrite=True
          )
          encoded_video = self._ffmpeg_decode(video_temp_path)
        finally:
          os.unlink(video_temp_path)
      else:
        encoded_video = self._ffmpeg_decode(video_or_path_or_fobj)
    elif isinstance(video_or_path_or_fobj, bytes):
      with tempfile.TemporaryDirectory() as tmpdirname:
        video_temp_path = os.path.join(tmpdirname, 'video')
        with tf.io.gfile.GFile(video_temp_path, 'wb') as f:
          f.write(video_or_path_or_fobj)
        encoded_video = self._ffmpeg_decode(video_temp_path)
    elif hasattr(video_or_path_or_fobj, 'read'):
      encoded_video = self._ffmpeg_decode(video_or_path_or_fobj)
    else:  # List of images, np.array,...
      encoded_video = video_or_path_or_fobj
    return super(Video, self).encode_example(encoded_video)

  @classmethod
  def from_json_content(
      cls, value: Union[Json, feature_pb2.VideoFeature]
  ) -> 'Video':
    if isinstance(value, dict):
      # For backwards compatibility
      shape = tuple(value['shape'])
      encoding_format = value['encoding_format']
      ffmpeg_extra_args = value['ffmpeg_extra_args']
      return cls(
          shape=shape,
          encoding_format=encoding_format,
          ffmpeg_extra_args=ffmpeg_extra_args,
      )
    return cls(
        shape=feature_lib.from_shape_proto(value.shape),
        dtype=feature_lib.dtype_from_str(value.dtype),
        encoding_format=value.encoding_format or None,
        use_colormap=value.use_colormap,
        ffmpeg_extra_args=value.ffmpeg_extra_args,
    )

  def to_json_content(self) -> feature_pb2.VideoFeature:  # pytype: disable=signature-mismatch  # overriding-return-type-checks
    return feature_pb2.VideoFeature(
        shape=feature_lib.to_shape_proto(self.shape),
        dtype=feature_lib.dtype_to_str(self.dtype),
        encoding_format=self._encoding_format,
        use_colormap=self._use_colormap,
        ffmpeg_extra_args=self._extra_ffmpeg_args,
    )

  def repr_html(self, ex: np.ndarray) -> str:
    """Video are displayed as `<video>`."""
    return image_feature.make_video_repr_html(
        ex, use_colormap=self.feature._use_colormap  # pylint: disable=protected-access  # pytype: disable=attribute-error
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""`tfds new` command."""

import argparse
import os
import pathlib
import subprocess
import textwrap
from typing import Optional

from tensorflow_datasets.core import constants
from tensorflow_datasets.core import dataset_metadata
from tensorflow_datasets.core import naming
from tensorflow_datasets.core.utils import resource_utils
from tensorflow_datasets.scripts.cli import builder_templates
from tensorflow_datasets.scripts.cli import cli_utils as utils


def register_subparser(parsers: argparse._SubParsersAction) -> None:  # pylint: disable=protected-access
  """Add subparser for `new` command."""
  new_parser = parsers.add_parser(
      'new', help='Creates a new dataset directory from the template.'
  )
  new_parser.add_argument(
      'dataset_name',  # Positional argument
      type=str,
      help='Name of the dataset to be created (in snake_case)',
  )
  new_parser.add_argument(
      '--data_format',  # Optional argument
      type=str,
      default=builder_templates.STANDARD,
      choices=[
          builder_templates.STANDARD,
          builder_templates.CONLL,
          builder_templates.CONLLU,
      ],
      help=(
          'Optional format of the input data, which is used to generate a '
          'format-specific template.'
      ),
  )
  new_parser.add_argument(
      '--dir',
      type=pathlib.Path,
      default=pathlib.Path.cwd(),
      help=(
          'Path where the dataset directory will be created. '
          'Defaults to current directory.'
      ),
  )
  new_parser.set_defaults(subparser_fn=_create_dataset_files)


def _create_dataset_files(args: argparse.Namespace) -> None:
  """Creates the dataset directory. Executed by `tfds new <name>`."""
  if not naming.is_valid_dataset_and_class_name(args.dataset_name):
    raise ValueError(
        'Invalid dataset name. It should be a valid Python class name.'
    )

  create_dataset_files(
      dataset_name=args.dataset_name,
      dataset_dir=args.dir,
      data_format=args.data_format,
  )


def create_dataset_files(
    dataset_name: str,
    dataset_dir: pathlib.Path,
    data_format: Optional[str] = None,
) -> None:
  """Creates the dataset files."""
  # Creates the root directory
  dataset_dir = dataset_dir.expanduser().resolve() / dataset_name
  dataset_dir.mkdir(parents=True)
  in_tfds = 'tensorflow_datasets' in dataset_dir.parts

  info = utils.DatasetInfo(
      name=dataset_name,
      in_tfds=in_tfds,
      path=dataset_dir,
      data_format=data_format,
  )

  _create_dataset_file(info)
  _create_dataset_test(info)
  _create_dataset_readme(info)
  _create_dataset_citations(info)
  _create_dataset_tags(info)
  _create_init(info)
  _create_dummy_data(info)
  _create_checksum(info)
  if in_tfds:
    _add_to_parent_init(info)

  print(
      'Dataset generated at {}\n'
      'You can start searching `{}` to complete the implementation.\n'
      'Please check '
      'https://www.tensorflow.org/datasets/add_dataset for additional details.'
      .format(info.path, info.todo)
  )


def _create_dataset_file(info: utils.DatasetInfo) -> None:
  """Create a new dataset from a template."""
  file_path = info.path / f'{info.name}_dataset_builder.py'

  content = builder_templates.create_builder_template(info)
  file_path.write_text(content)


def _create_dataset_test(info: utils.DatasetInfo) -> None:
  """Adds the `dummy_data/` directory."""
  file_path = info.path.joinpath(f'{info.name}_dataset_builder_test.py')

  content = textwrap.dedent(f'''\
      """{info.name} dataset."""

      from {info.ds_import} import {info.name}_dataset_builder
      import {info.tfds_api} as tfds

      class {info.cls_name}Test(tfds.testing.DatasetBuilderTestCase):
        """Tests for {info.name} dataset."""
        # {info.todo}:
        DATASET_CLASS = {info.name}_dataset_builder.Builder
        SPLITS = {{
            'train': 3,  # Number of fake train example
            'test': 1,  # Number of fake test example
        }}

        # If you are calling `download/download_and_extract` with a dict, like:
        #   dl_manager.download({{'some_key': 'http://a.org/out.txt', ...}})
        # then the tests needs to provide the fake output paths relative to the
        # fake data directory
        # DL_EXTRACT_RESULT = {{'some_key': 'output_file1.txt', ...}}


      if __name__ == '__main__':
        tfds.testing.test_main()
      ''')
  file_path.write_text(content)


def _create_dataset_readme(info: utils.DatasetInfo) -> None:
  """Adds the `README.md` file."""
  file_path = info.path / 'README.md'
  content = textwrap.dedent(f"""\
      {info.todo}: Markdown description of that will appear on the catalog page.
      Description is **formatted** as markdown.

      It should also contain any processing which has been applied (if any),
      (e.g. corrupted example skipped, images cropped,...):
      """)
  file_path.write_text(content)


def _create_dataset_citations(info: utils.DatasetInfo) -> None:
  """Adds the `CITATIONS.bib` file."""
  file_path = info.path / 'CITATIONS.bib'
  content = textwrap.dedent(f"""\
      // {info.todo}: BibTeX citation
      """)
  file_path.write_text(content)


def _create_dataset_tags(info: utils.DatasetInfo) -> None:
  """Adds the `TAGS.txt` file."""
  file_path = info.path / 'TAGS.txt'
  content = '// {info.todo}: remove tags which do not apply to dataset.\n' + (
      dataset_metadata.valid_tags_with_comments()
  )
  file_path.write_text(content)


def _create_init(info: utils.DatasetInfo) -> None:
  """Adds the `__init__.py` file."""
  file_path = info.path / '__init__.py'
  file_path.write_text('')


def _create_dummy_data(info: utils.DatasetInfo) -> None:
  """Adds the `dummy_data/` directory."""
  # Create a dummy file in the directory to force the directory creation in
  # version control system where empty directories aren't detected.
  file_path = info.path / 'dummy_data/TODO-add_fake_data_in_this_directory.txt'
  file_path.parent.mkdir()
  file_path.touch()


def _create_checksum(info: utils.DatasetInfo) -> None:
  """Adds the `checksums.tsv` file."""
  file_path = info.path / 'checksums.tsv'
  content = textwrap.dedent(f"""\
      # {info.todo}: If your dataset downloads files, then the checksums
      # will be automatically added here when running
      # `tfds build --register_checksums`.
      """)
  file_path.write_text(content)


def _add_to_parent_init(info: utils.DatasetInfo) -> None:
  """Add `import` dataset in the `<ds_type>/__init__.py` file."""
  if not info.in_tfds:
    # Could add global init, but would be tricky to foresee all use-cases
    raise NotImplementedError(
        'Adding __init__.py in non-tfds dir not supported.'
    )

  import_path = info.path.parent / '__init__.py'
  if not import_path.exists():
    return  # Should this be an error instead ?
  import_path.write_text(
      import_path.read_text()
      + f'from {info.ds_import} import {info.cls_name}\n'
  )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Wrapper around `tfds build`."""

import argparse
from typing import List

from absl import app
from absl import flags
from absl import logging

from tensorflow_datasets.scripts.cli import main as main_cli

module_import = flags.DEFINE_string('module_import', None, '`--imports` flag.')
dataset = flags.DEFINE_string('dataset', None, 'singleton `--datasets` flag.')

builder_config_id = flags.DEFINE_integer(
    'builder_config_id', None, '`--config_idx` flag'
)



def _parse_flags(argv: List[str]) -> argparse.Namespace:
  """Command lines flag parsing."""
  return main_cli._parse_flags([argv[0], 'build'] + argv[1:])  # pylint: disable=protected-access


_display_warning = True


def main(args: argparse.Namespace) -> None:
  if _display_warning:
    logging.warning(
        '***`tfds build` should be used instead of `download_and_prepare`.***'
    )
  if module_import.value:
    args.imports = module_import.value
  if dataset.value:
    args.datasets = [dataset.value]
  if builder_config_id.value is not None:
    args.config_idx = builder_config_id.value
  main_cli.main(args)


if __name__ == '__main__':
  app.run(main, flags_parser=_parse_flags)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Historical phenological data for cherry tree flowering at Kyoto City."""

from __future__ import annotations

import csv

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

URL = 'https://raw.githubusercontent.com/rmcelreath/rethinking/master/data/cherry_blossoms.csv'

_DESCRIPTION = """
Historical phenological data for cherry tree flowering at Kyoto City.

This data was collected from diaries and chronicles dating back to the 9th
century. Data from the 9th to the 14th century were collected by Aono and Saito
(2010;  International Journal of Biometeorology, 54, 211-219), while the 15th
to 21st  century were collected by Aono and Kazui (2008; International Journal
of  Climatology, 28, 905-914).

All dates are expressed in the Gregorian calendar.


Number of instances: 1216

Variables:

1. year: Year CE  (int)
2. doy: Day of year of first bloom. Day 89 is April 1. Day 119 is May 1. (float)
3. temp: March temperature estimate (float)
4. temp_upper: Upper 95% bound for estimate (float)
5. temp_lower: Lower 95% bound for estimate (float)
"""

_CITATION = """
@ONLINE {
    author = "Aono, Yasuyuki",
    title  = "Historical Series of Phenological data for Cherry Tree Flowering at Kyoto City (and March Mean Temperature Reconstructions)",
    year   = "2012",
    url    = "http://atmenv.envi.osakafu-u.ac.jp/aono/kyophenotemp4/"
}
"""


class CherryBlossoms(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for cherry_blossoms dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    features_dict = {
        'year': tf.int32,
        'doy': tf.float32,
        'temp': tf.float32,
        'temp_upper': tf.float32,
        'temp_lower': tf.float32,
    }
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict(features_dict),
        supervised_keys=None,  # Set to `None` to disable
        homepage='http://atmenv.envi.osakafu-u.ac.jp/aono/kyophenotemp4/',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    path = dl_manager.download_and_extract(URL)
    return {tfds.Split.TRAIN: self._generate_examples(path)}

  def _generate_examples(self, path):
    """Yields examples."""
    with path.open() as f:  # pytype: disable=attribute-error  # gen-stub-imports
      reader = csv.DictReader(f, delimiter=';')
      for index, row in enumerate(reader):
        example = {k: v.replace('NA', 'nan') for k, v in row.items()}
        yield index, example



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=line-too-long
"""`tensorflow_datasets` (`tfds`) defines a collection of datasets ready-to-use with TensorFlow.

Each dataset is defined as a `tfds.core.DatasetBuilder`, which encapsulates
the logic to download the dataset and construct an input pipeline, as well as
contains the dataset documentation (version, splits, number of examples, etc.).

The main library entrypoints are:

* `tfds.builder`: fetch a `tfds.core.DatasetBuilder` by name
* `tfds.load`: convenience method to construct a builder, download the data, and
  create an input pipeline, returning a `tf.data.Dataset`.

Documentation:

* These API docs
* [Available datasets](https://www.tensorflow.org/datasets/catalog/overview)
* [Colab tutorial](https://colab.research.google.com/github/tensorflow/datasets/blob/master/docs/overview.ipynb)
* [Add a dataset](https://www.tensorflow.org/datasets/add_dataset)
"""
# pylint: enable=line-too-long
# pylint: disable=g-import-not-at-top,g-bad-import-order,wrong-import-position,unused-import

import time

_TIMESTAMP_IMPORT_STARTS = time.time()
from absl import logging
import tensorflow_datasets.core.logging as _tfds_logging
from tensorflow_datasets.core.logging import call_metadata as _call_metadata

_metadata = _call_metadata.CallMetadata()
_metadata.start_time_micros = int(_TIMESTAMP_IMPORT_STARTS * 1e6)
_import_time_ms_dataset_builders = 0

try:
  # Imports for registration
  _before_dataset_imports = time.time()
  from tensorflow_datasets import dataset_collections

  # pytype: disable=import-error
  # For builds that don't include all dataset builders, we don't want to fail on
  # import errors of dataset builders.
  try:
    from tensorflow_datasets import audio
    from tensorflow_datasets import graphs
    from tensorflow_datasets import image
    from tensorflow_datasets import image_classification
    from tensorflow_datasets import object_detection
    from tensorflow_datasets import nearest_neighbors
    from tensorflow_datasets import question_answering
    from tensorflow_datasets import d4rl
    from tensorflow_datasets import ranking
    from tensorflow_datasets import recommendation
    from tensorflow_datasets import rl_unplugged
    from tensorflow_datasets import rlds
    from tensorflow_datasets import robotics
    from tensorflow_datasets import robomimic
    from tensorflow_datasets import structured
    from tensorflow_datasets import summarization
    from tensorflow_datasets import text
    from tensorflow_datasets import text_simplification
    from tensorflow_datasets import time_series
    from tensorflow_datasets import translate
    from tensorflow_datasets import video
    from tensorflow_datasets import vision_language

  except ImportError:
    pass
  # pytype: enable=import-error

  _import_time_ms_dataset_builders = int(
      (time.time() - _before_dataset_imports) * 1000
  )

  # Public API to create and generate a dataset
  from tensorflow_datasets.public_api import *  # pylint: disable=wildcard-import
  from tensorflow_datasets import public_api  # pylint: disable=g-bad-import-order
  # __all__ for import * as well as documentation
  __all__ = public_api.__all__

except Exception as exception:  # pylint: disable=broad-except
  _metadata.mark_error()
  logging.exception(exception)
finally:
  _metadata.mark_end()
  _tfds_logging.tfds_import(
      metadata=_metadata,
      import_time_ms_tensorflow=0,
      import_time_ms_dataset_builders=_import_time_ms_dataset_builders,
  )
